%PROJECTIONDATA stores CT projections, scan metadata, and reconstructions
% the PROJECTIONDATA class defines objects for storing CT projection data,
% metadata associated with these projections, and reconstructions. It is
% currently built specifically for data from the Rigaku Nano3DX scanner.
% For an overview of properties (fields) of PROJECTIONDATA objects, call:
%   properties(projectionData)
% For an overview of methods (functions) of PROJECTIONDATA objects, call:
%   methods(projectionData)
% Alternatively, you can also call METHODS and PROPERTIES with an object of
% of the given class as input.
% To get information on a specific METHOD or PROPERTY, call:
%   help projectionData.METHOD
%   help projectionData.PROPERTY
% Again, you can also use this syntax with the name of an object of this
% class instead of the name of the class
% To get more information on the construction of objects of the
% PROJECTIONDATA class, call:
%   help projectionData.projectionData

% TODO - rename to n3DXData?
classdef projectionData < matlab.mixin.Copyable % subclass of handle class, allows creating copies of object
    properties
        % 3D array of projection data (type single). projection data are
        % oriented so that one X-Y slice of the volume is one projection,
        % and projections are right side up.
        proj(:,:,:) single;
        % 3D array of reconstructed data (type single)
        tomo(:,:,:) single;
        % structure for any info that doesn't fit into other parameters
        misc struct = struct;
    end
    properties (SetAccess = protected)
        % (write access only through class methods) object containing
        % information about scan geometry. this information is useful for
        % reconstruction, but also for processing functions that require
        % some information about the scan geometry. Other functions may
        % also alter information contained in geom through the GEOM.EDIT
        % method. For more information, call: help scanGeometry
        geom scanGeometry = scanGeometry;
        % (read-only) structure of scan metadata loaded from a *.ctm file
        % associated with the projection data. 
        params struct = struct;
        % (write only through class methods) object containing
        % settings such as input/output paths, and names of processing and
        % reconstruction functions. May be edited using its class methods.
        % For more information, call: help recSettings
        settings recSettings = recSettings;
        % (read-only) property cointaining a log of operations done with
        % the object
        log char = '';
    end
    
    methods
        function obj = projectionData(inputPath, previewFlag) % constructor
            %Initialize PROJECTIONDATA object (constructor method)
            % this is the constructor method for an object of class
            % projectionData. When called with no inputs, an empty object
            % of this class is created. Alternatively, an input path
            % character vector can be passed as an input to create a
            % projectionData object and immediately load the projections
            % and metadata referenced by the path
            %   Inputs:
            %       previewFlag - logical flag whether to load the entire
            %       dataset, or only the sinogram of its central slice
            %       (optional, defaults to false)
            %   Outputs:
            %       obj - projectionData object
            %   Example call:
            %       data = projectionData('C:\placeholdername.raw') % whole dataset
            %       data = projectionData('C:\placeholdername.raw', false) % whole dataset
            %       data = projectionData('C:\placeholdername.raw', 1) % only preview
            
            % TODO - is it necessary?
            obj.settings = recSettings;
            obj.geom = scanGeometry;
            obj.geom.changeParent(obj);
            
            if nargin > 0 % if a recSet object was passed to the constructor
                if nargin < 2 % if previewFlag wasn't supplied
                    previewFlag = false;
                end % TODO - validate previewflag?
                
                % add input to object
                obj.settings.addInput(inputPath);
                
                % when an input path is supplied 
                obj.loadProj(previewFlag);
            end
            
            % NB: obj.settings and obj.geom are instances of other handle
            % classes, when delete(obj) is called, matlab should also
            % implicitly call the delete methods of obj.settings and
            % obj.geom
        end
        
        function loadProj(obj, previewFlag)
            %LOADPROJ - load projections, metadata into object properties
            % the loadProj method takes as input the recSet object and 
            % loads the projections as well as the metadata referenced in
            % recSet. Projections are saved to the proj property, metadata
            % into the params property. Parameters of the scan geometry are
            % calculated from the loaded projections and metadata, and
            % saved in the geom property.
            % (it internally calls the loadMeta and extractGeom methods).
            %
            %   Inputs:
            %       recSet - recSettings object
            %       previewFlag - logical flag denoting whether to load the
            %           entire dataset or just the middle sinogram
            %           (default - false)
            %   Example call to load full dataset:
            %       data.loadProj(recSet, 0) or data.loadProj(recSet)
            %   Example call to load preview:
            %       data.loadProj(recSet, 1)
            
            % TODO - validate previewFlag (scalar etc)
            
            % if previewFlag is not supplied, set it to false by default
            if nargin < 2
                previewFlag = false;
            end
            
            % load metadata
            obj.loadMeta;
            
            % calculate scan geometry parameters from the loaded ctm file
            obj.geom.extractGeom;
            
            % get path to projection file
            filename = [obj.settings.input.path, ...
                        obj.settings.input.name, ...
                        obj.settings.input.format];
            
            % create a memory map to the projection file
            projMap = memmapfile(filename, ...
                'Format', {'uint16', ...
                [obj.geom.width, obj.geom.height, obj.geom.num_proj], 'x'});
            
            % load the data or preview
            if previewFlag
                % get index of middle row
                middle = round( (obj.geom.height + 1)/2);
                % use the file map to get the middle row
                tempProj = projMap.Data.x(:, middle, :);
                
                % only a single row - height will be 1
                obj.geom.edit('height', 1)
                obj.geom.edit('reconHeight', 1)
                
                % if loading a preview, also load 1st projection into misc
                firstProj = projMap.Data.x(:, :, 1);
                % permute and convert to single
                firstProj = single( permute(firstProj, [2 1]) );
                obj.misc.firstProj = firstProj;
            else
                tempProj = projMap.Data.x;
            end
            % right now tempProj only points to projMap.Data.x
            % only once the memmapfile object is cleared, data is actually
            % written into RAM, into the tempProj variable
            clear projMap
            
            % Permute projections, so they are right side up
            tempProj = permute(tempProj, [2 1 3]);
            % same for full data as well as preview
            
            % Convert tempProj to single - most demanding operation
            tempProj = single(tempProj);
            
            % NB: 3 most demanding operations are clearing projMap (data
            % written into RAM at that point), conversion to single, and
            % array permutation (projections look better after permuting)
            
            % finally, write projections into the object's proj property
            obj.proj =  tempProj;
            % save into misc info about whether it is only a preview or not
            obj.misc(1).preview = logical(previewFlag);
        end
        % TODO - add support for more filetypes?
        % TODO - can data be a type other than uint16? if yes, I will have to account for that
        
        function loadMeta(obj)
            %LOADMETA attempts to load info from metafile into PARAMS
            % This method assumes the metadata file is an XML file,
            % preferably with the suffix *.ctm
            
            % TODO --- Check Inputs
            
            % get path to metadata file
            metafilename = [obj.settings.input.metapath, ...
                            obj.settings.input.metaname, ...
                            obj.settings.input.metaformat];
            % read *.ctm file, convert to Matlab structure
            file = xml2struct(metafilename);
            % get only part about scan metadata
            if strcmp(obj.settings.input.metaformat, '.ctm')
                metadata = file.nano3DX.CT_measurement;
            elseif strcmp(obj.settings.input.metaformat, '.ipro') % TODO - make this if more elegant
                metadata = file.nano3DX.Projection_measurement;
            end
            % get field names in metadata structure
            fields = fieldnames(metadata);
            % go entry by entry and check if they are numbers or not
            for k = 1:length(fields)
                % get current field's value and convert to double
                pom = str2double(metadata.(fields{k}).Text);
                if isnan(pom) % if str2double returns NaN, write current field as string
                    % check if it is a logical value
                    if strcmpi(metadata.(fields{k}).Text, 'TRUE') % case insensitive
                        obj.params(1).(fields{k}) = true;
                    elseif strcmpi(metadata.(fields{k}).Text, 'FALSE')
                        obj.params(1).(fields{k}) = false;
                    else % else just write the text
                        obj.params(1).(fields{k}) = metadata.(fields{k}).Text;
                    end
                else % else write as double
                    obj.params.(fields{k}) = pom;
                end
            end
        end
        
        
        function sinogram = getSino(obj, slices, avgFlag)
            % create another projectionData object with only one sinogram
            % if a specific slice is not chosen, the middle slice is taken
            % slices can either be a scalar or a vector of the first and
            % last slice - in the second case, if avgFlag is true, the 
            % output is an average of the chosen range. If not, it is a 3D
            % array of individual sinograms
            % TODO - update description
            % TODO - check that slices has a proper format (integer, ascending, only length 1 or 2...)
            
            if nargin < 3
                avgFlag = false;
            end
            
            % create a new projectionData object
            sinogram = projectionData;
            % copy all other properties from obj
            sinogram.geom = copy(obj.geom);
            sinogram.geom.changeParent(sinogram); % TODO - make this better
            
            % write sinogram into proj property of new object
            if nargin == 1 % if no slice is given
                slices = round( (obj.geom.height + 1)/2); % get index of middle
            else
                % get a vector of slice indexes
                slices = round( (slices(1)):(slices(end)) ); 
                % make sure that these indexes aren't out of bounds
                slices(slices < 1) = [];
                slices(slices > obj.geom.height) = [];
            end% TODO - throw warning about this?

                sinogram.params = obj.params;
                sinogram.settings = copy(obj.settings);
                sinogram.misc = obj.misc;
                
                % get the desired sinograms
                sinogramRange = obj.proj(slices, :, :);
                
                if avgFlag 
                    sinogramRange = sum(sinogramRange, 1) / size(sinogramRange, 1); % average all projection rows
                end
                
                sinogram.proj = sinogramRange;
                
                % update height of sinogram
                sinogram.geom.edit('height', size(sinogram.proj, 1)); % TODO - remove if automatic dimension checking is implemented in projectionData
                sinogram.geom.edit('reconHeight', size(sinogram.proj, 1));
        end
        
        function sinogram = getSinoArray(obj, slices, avgFlag)
            % create an array (not another projectionData object) with 
            % only one sinogram. If a specific slice is not chosen, the 
            % middle slice is taken. Slices can either be a scalar or a 
            % vector of the first and last slice - in the second case, if 
            % avgFlag is true, the  output is an average of the chosen 
            % range. If not, it is a 3D array of individual sinograms
            % TODO - better description
            % TODO - check that slices has a proper format (integer, ascending, only length 1 or 2...)
            
            % TODO - check avgFlag
            if nargin < 3
                avgFlag = false;
            end
            
            % write sinogram an array
            if nargin == 1 % if no slice is given
                middle = round( (obj.geom.height + 1)/2); % get index of middle
                sinogram = obj.proj(middle, :, :); % turn into 2D
            else
                
                % get a vector of slice indexes
                slices = round( (slices(1)):(slices(end)) );
                % make sure that these indexes aren't out of bounds
                slices(slices < 1) = [];
                slices(slices > obj.geom.height) = [];
                % TODO - throw warning about this?
                % get the desired sinograms
                sinogram = obj.proj(slices, :, :);
                
                if avgFlag 
                    sinogram = sum(sinogram, 1) / size(sinogram, 1); % average all projection rows
                end
                
            end
        end
        
        
        function bytes = getProjBytes(obj)
            % output size of projection data in bytes
            
            % extract proj to new variable. Pretty much no overhead, matlab uses copy-on-write
            tempProj = obj.proj;
            % get information on projection array using whos
            s = whos('tempProj');
            % output the bytes field
            bytes = s.bytes;
        end
        
        
        function bytes = getTomoBytes(obj)
            % output size of tomogram data in bytes
            
            % extract tomo to new variable. Pretty much no overhead, matlab uses copy-on-write
            tempTomo = obj.tomo;
            % get information on tomogram array using whos
            s = whos('tempTomo');
            % output the bytes field
            bytes = s.bytes;
        end
        
        % TODO - update geom height width and num_proj every time
        % projections are overwritten?
        
        % TODO - add subsref, subsasgn, get/set methods
        
        % TODO - add destructor method, study up on how object is destroyed
        % delete(data) % to delete the data object itself (clear data just deletes the handle)
        
        
        function editParams(obj, varargin)
            % Allows for editing information in the PARAMS property, which
            % contains scan metadata. This is usually extracted from the
            % *.ctm file using the LOADMETA method, but this function may
            % be used for debugging/experimenting purposes.
            
            p = inputParser; 
            p.FunctionName = 'editParams';
            p.KeepUnmatched = true;
            % parse
            parse(p, varargin{:})
            
            put_in_params = p.Unmatched;
            
            fields = fieldnames(put_in_params);
            
            for k = 1:numel(fields)
                obj.params.(fields{k}) = put_in_params.(fields{k});
            end
            
        end
        
        
        function clearParams(obj)
            % Allows for clearing information in the PARAMS property, which
            % contains scan metadata. This function maybe used for 
            % debugging/experimenting purposes.
            obj.params = struct;
        end
        
        function writeToLog(obj, logEntry)
            % TODO - update description
            obj.log = sprintf('%s\n%s - %s', obj.log, ...
                                              datestr(now), ...
                                              logEntry);
        end
        
        
        
        
    end

    methods(Access = protected, Hidden = true)
      function copied = copyElement(obj) % override default copy behavior
          % to copy attributes that are handle objects - recSettings and
          % projGeometry (otherwise only their handle is copied)
          copied = projectionData;
          copied.proj = obj.proj;
          copied.tomo = obj.tomo;
          copied.misc = obj.misc;
          
          copied.geom = copy(obj.geom);
          copied.geom.changeParent(copied);
          
          copied.params = obj.params;
          copied.settings = copy(obj.settings);
          copied.log = obj.log;
          
          writeToLog(copied, 'Copied to a different object')
      end
   end
    
end

