%SCANGEOMETRY - object for storing scan geometry

% TODO - update description, add list of methods...

classdef scanGeometry < matlab.mixin.Copyable
    properties
        % width of projections, in pixels (whole, non-negative scalar)
        width(1,1) double {mustBeNumeric, mustBeInteger, mustBeNonnegative, mustBeFinite} = 0;
        % height of projections, in pixels (whole, non-negative scalar)
        height(1,1) double {mustBeNumeric, mustBeInteger, mustBeNonnegative, mustBeFinite} = 0;
        % length of the edge of one pixel (positive, finite scalar)
        % NB: voxels are assumed isotropic
        pix_size(1,1) double {mustBeNumeric, mustBePositive, mustBeFinite} = 1;
        % Source-origin distance, in micrometers (non-negative)
        SOD(1,1) double {mustBeNumeric, mustBeNonnegative} = inf;
        % Origin-detector distance, in micrometers (non-negative)
        ODD(1,1) double {mustBeNumeric, mustBeNonnegative} = 0;
        % Vector of angles, under which individual projections were taken
        angles(1,:) double {mustBeNumeric, mustBeFinite} = [];
        % Location of the center of rotation as the index of detector
        % pixels in the horizontal direction (starting from the left, does
        % not have to be a whole number)
        CoR(1,1) double {mustBeNumeric, mustBeFinite} = 0.5;
        % Width of resulting reconstruction volume (in pixels, does not
        % have to equal projection width, reconstructed tomograms will be
        % square)
        reconWidth(1,1) double {mustBeNumeric, mustBeInteger, mustBeNonnegative, mustBeFinite} = 0;
        % Height of resulting reconstruction volume (in pixels, does not
        % have to equal projection height)
        reconHeight(1,1) double {mustBeNumeric, mustBeInteger, mustBeNonnegative, mustBeFinite} = 0;
        % Logical flag denoting whether the CT scan was a step scan (true),
        % or continuous scan (false)
        stepscanFlag(1,1) logical = false;
        % Logical flag denoting whether the sample was offset to one side
        % during the CT scan to extend the field of view
        offsetFlag(1,1) logical = false;
        % Logical flag denoting whether the CT scan was a shortscan
        % (between 180 and 360) or not (usually a half scan or full scan)
        shortscanFlag(1,1) logical = false;
        % TODO - add x_shift, y_shift?
        
        % TODO - make num_proj dependent on numel(angles) ?
        % TODO - add help describing various fields in geom
        % TODO - check class property validation again
    end
    properties (Dependent)
        % Shift of the center of rotation away from the middle of
        % projections (in pixels)
        CoR_shift
    end
    properties (Dependent, SetAccess = private)
        % number of projections (read-only, depends on the angles property)
        num_proj
        % source-detector distance, in micrometers (read-only, depends on
        % SOD and ODD)
        SDD
        % cone beam magnification (read-only, depends on SOD and ODD)
        cb_mag
        % voxel size of the reconstructed volume, in micrometers
        % (read-only, depends on SOD, ODD, and pixel size at the detector)
        vox_size
    end
    properties (Hidden = true, SetAccess = private)
        parent % parent PROJECTIONDATA object % TODO - do this better?
    end
    
    methods (Access = ?projectionData) % only constructible by a projectionData object
        
        function obj = scanGeometry % constructor
            %Initialize 
            
        end
        
        function changeParent(obj, data) % in case I need to change the object's parent
            %Initialize 
            
            obj.parent = data;
            
        end

    end
    
    methods
        
        function extractGeom(obj)
            %EXTRACTGEOM extracts scan geometry from proj and params.
            % numElements is the total number of pixels in projection data,
            % and it is needed to determine whether the scan was a step or
            % continuous scan
            
            recSet = obj.parent.settings;
            
            % get the full filename, including path
            filename = [recSet.input.path, ...
                        recSet.input.name, ...
                        recSet.input.format];
            % get number of pixels in file
            fileinfo = dir(filename); % get information about file;
            numElements = fileinfo.bytes / 2; % /2 'cause 16 bits (2 bytes)
            
            parms = obj.parent.params; % write params into temporary variable
            % number of detector columns
            obj.width = parms.roi_dx_pixel;
            % number of detector rows
            obj.height = parms.roi_dy_pixel;
            % number of projections (temporary variable)
            no_proj = numElements / (obj.width * obj.height);
            
            % determine if step scan based on projection number
            if no_proj == parms.num_projections
                obj.stepscanFlag = false;
            else
                obj.stepscanFlag = true;
                % (has 2 more projections than declared in ctm file)
            end
            
            % flag whether it is offset scan or not
            if isfield(parms, 'offsetscan') && parms.offsetscan == 1
                obj.offsetFlag = true;
            else
                obj.offsetFlag = false;
            end
            
            % actual pixel size when taking binning into account
            obj.pix_size = parms.xd_pixelsize * parms.binning;
            % source-object distance, in um
            obj.SOD = parms.PixelSizeFactor_L * 1000;
            % object-detector distance, in um
            obj.ODD = (parms.PixelSizeFactor_dL + parms.XD_position) * 1000;
            
            % get projection angles in radians. slightly different for step and continuous scans
            firstAngle = parms.start_angle; % in degrees
            lastAngle = parms.start_angle + parms.angle_range; % in degrees
            if obj.stepscanFlag % stepscan has 2 extra projections, 1st and last, both at parms.angle_range
                if obj.offsetFlag
                    temp_angles = linspace(firstAngle, lastAngle, no_proj);
                else
                    % get all angles (still in degrees)
                    temp_angles = linspace(firstAngle, lastAngle, no_proj-1);
                    temp_angles(end) = [];
                    % add the 2 extra projections for step scan
                    temp_angles = [lastAngle, temp_angles, lastAngle];
                end
            else % continuous scan just has the regular scan range, no more
                % get all angles (in degrees)
                temp_angles = linspace(firstAngle, lastAngle, no_proj+1);
                temp_angles(end) = [];
            end
            temp_angles = temp_angles .* pi/180; % finally convert to radians
            obj.angles = temp_angles;
            % Center of Rotation - assumed in the middle by default
            obj.CoR = (obj.width+1)/2;
            % width of reconstruction volume (slices will always be squares);
            obj.reconWidth = parms.roi_dx_pixel;
            % height of reconstruction volume (number of slices);
            obj.reconHeight = parms.roi_dy_pixel;
            % flag whether scan is shortscan or not
            if parms.angle_range > 180 && parms.angle_range < 360
                obj.shortscanFlag = true;
            else
                obj.shortscanFlag = false;
            end
            
        end
        
        function edit(obj, varargin)
            %EDIT takes as input the scanGeometry object obj and edits it
            % according to the rest of the inputs. The function will accept
            % inputs that do not correspond with fields in geom, but these
            % inputs will not affect geom in any way.
            %   Inputs:
            %       obj - scanGeometry object
            %       Name-Value pairs:
            %           1) name of a property, e.g. 'CoR'
            %           2) value to write into the given field, e.g. 100
            %           For a list of properties, call:
            %               properties(scanGeometry)
            % 
            % obj = obj.edit(varargin);
            
            % --- Check and parse inputs ---
            % set up parser object and modify its settings
            p = inputParser; 
            p.FunctionName = 'edit';
            p.KeepUnmatched = true;
            % check additional name-value pair inputs
            addParameter(p, 'width',         NaN, ...
                @(x) isnumeric(x) && isscalar(x) && (floor(x)==x))
            addParameter(p, 'height',        NaN, ...
                @(x) isnumeric(x) && isscalar(x) && (floor(x)==x))
            addParameter(p, 'num_proj',      NaN, ...
                @(x) isnumeric(x) && isscalar(x))
            addParameter(p, 'pix_size',      NaN, ...
                @(x) isnumeric(x) && isscalar(x))
            addParameter(p, 'SOD',           NaN, ...
                @(x) isnumeric(x) && isscalar(x))
            addParameter(p, 'ODD',           NaN, ...
                @(x) isnumeric(x) && isscalar(x))
            addParameter(p, 'angles',        NaN, ...
                @(x) isnumeric(x) && isvector(x))
            addParameter(p, 'CoR',           NaN, ...
                @(x) isnumeric(x) && isscalar(x))
            addParameter(p, 'CoR_shift',     NaN, ...
                @(x) isnumeric(x) && isscalar(x))
            addParameter(p, 'reconWidth',    NaN, ...
                @(x) isnumeric(x) && isscalar(x) )
            addParameter(p, 'reconHeight',   NaN, ...
                @(x) isnumeric(x) && isscalar(x) )
            addParameter(p, 'stepscanFlag',  NaN, ...
                @(x) (isnumeric(x) || islogical(x)) && isscalar(x) )
            addParameter(p, 'offsetFlag',    NaN, ...
                @(x) (isnumeric(x) || islogical(x)) && isscalar(x) )
            addParameter(p, 'shortscanFlag', NaN, ...
                @(x) (isnumeric(x) || islogical(x)) && isscalar(x) )
            % parse
            parse(p, varargin{:})
            % Go through inputs. If non-NaN, add to/overwrite in, obj.geom
            % go entry by entry and check if they are numbers or not
            for k = 1:length(p.Parameters) % p.Parameters contains parameter names
                if ~isnan(p.Results.(p.Parameters{k}))
                    obj.(p.Parameters{k}) = p.Results.(p.Parameters{k});
                end
            end
        end
        % TODO - check all these conditions, add what is necessary, throw errors
        % TODO - maybe this can be done differently now?
        
        function cb_mag = get.cb_mag(obj)
            % get cone beam magnification
            if obj.SOD == inf
                cb_mag = 1; % for infinite SOD, just set cone beam magnification straight to 1
            else
                cb_mag = (obj.SOD + obj.ODD) / obj.SOD;
            end
        end
        
        function vox_size = get.vox_size(obj)
            % get voxel size (in um)
            vox_size = obj.pix_size / obj.cb_mag;
            % NB: voxel size is in the ctm file, too (as <PixelSize>
        end
        
        function CoR_shift = get.CoR_shift(obj)
            % get axis-of-rotation shift from middle of detector (in pixels)
            CoR_shift = obj.CoR - (obj.width + 1)/2;
            % width of projections - detector bin at CoR
        end
        
        function set.CoR_shift(obj, CoR_shift)
            % set axis-of-rotation shift from middle of detector (in pixels)
            obj.CoR = (obj.width + 1)/2 + CoR_shift;
            % width of projections minus detector bin at CoR
        end
        
        function SDD = get.SDD(obj)
            % get source-detector distance
            SDD = obj.SOD + obj.ODD;
        end
        
        function num_proj = get.num_proj(obj)
            % get number of projections
            num_proj = numel(obj.angles);
        end
        
        % TODO - update geom height width and num_proj every time
        % projections are overwritten?
        
        % TODO - add subsref, subsasgn, get/set methods
        
    end
   
end

