function [taskLog] = taskRunner(data, tasksToRun)
%TAKSRUNNER takes functions and inputs written in data.settings and runs them
%	TASKRUNNER takes names of functions in the TASKS property of the
%	RECSETTINGS object DATA.SETTINGS, checks whether the function 
%   exists in the current path, and if it is, attempts running the function
%   with DATA (1st input) as the first input, and any additional inputs, if
%   any are saved in DATA.SETIINGS. the third input, TASKSTORUN, specifies which
%   functions saved in DATA.SETIINGS should be run.
%   NB: if the current task is 'save', taskRunner will attempt to save the
%   projections instead of looking for a function in path
% Inputs:
%   data - PROJECTIONDATA object, containing projection data in the PROJ
%   property
%   tasksToRun - optional input specifying what tasks to run. If left empty
%   or set to 'all', all functions specified in DATA.SETTINGS.TASKS are run.
%   Otherwise, it can be a numeric vector specifying the indexes of
%   functions that are to be run.
% Outputs:
%   taskLog - log variable containing information about the tasks that have
%   been run.
% Example:
%   taskLog = taskRunner(data);
%   taskLog = taskRunner(data, 'all');
%   taskLog = taskRunner(data, 1);
%   taskLog = taskRunner(data, [1 2 3]);


% how many tasks are listed in the recSettings object
totalTaskNumber = length(data.settings.tasks);

if nargin < 2 % if taskNumbers was not supplied
    % get indexes of all tasks - in order
    tasksToRun = 1:totalTaskNumber;
elseif isequal(lower(tasksToRun), 'all') % if it is a string saying 'all'
    % get indexes of all tasks - in order
    tasksToRun = 1:totalTaskNumber;
end
% TODO - validate taskNumbers (numeric, whole, upper/lower bound)

% create taskLog
taskLog = '';

% for loop going through all selected tasks
for k = 1:numel(tasksToRun)
    singleTaskStopwatch = tic;
    
    try
    
        if strcmp(data.settings.tasks{ tasksToRun(k) }.name, 'save')
            taskLog = sprintf('%s%s\n', taskLog, ...
                              'Attempting to save projections');
            % attempt saving projections (if they are to be saved)
            if data.settings.output.projFlag
                saveProjections(data)
            else
                taskLog = sprintf('%s%s\n', taskLog, ...
                    'Projection save flag set to false in data.settings. Projections were not saved.');
            end
            
        else % try finding the given task and running it

        % get the function name
        funcName = data.settings.tasks{ tasksToRun(k) }.name;

        % write into log
        taskLog = sprintf('%s%s %i: %s\n', taskLog, 'Attempting to run task', ...
                          tasksToRun(k), funcName);

            % get the structure containing inputs
            inputs = data.settings.tasks{ tasksToRun(k) }.inputs;

            % write inputs into log (if there are any)
            if ~isempty(fieldnames(inputs))
                % make cell structure w/ input names and values
                a = [fieldnames(inputs), struct2cell(inputs)];
                % use num2str on all cells
                a = cellfun(@num2str, a, 'UniformOutput', false);
                taskLog = sprintf('%s%s\n', taskLog, 'Inputs:');
                for m = 1:size(a, 1) % write every input into log
                    taskLog = sprintf('%s\t%s: %s\n', taskLog, a{m,1}, a{m,2});
                end
            end

            % search in path for the m-file associated with the function
            funcPath = which(funcName);
            
            if ~isempty(funcPath) % if the function is associated with a file,
                % get a handle to the function
                funcHandle = str2func(funcName);
                % get first line of the function
                a = fopen(funcPath);
                firstLine = fgetl( a );
                fclose(a);
                % find part in brackets (inputs), through regular expression
                [inputStart, inputEnd] = regexp(firstLine, '\(.*\)');
                % get a comma-separated list of inputs
                inputNames = firstLine( (inputStart+1):(inputEnd-1) );
                % turn comma-separated list into cell array
                inputNames = split(inputNames, {', ', ','}); % delimiters are either a comma followed by a space, or just a comma
                % 1st input should be the projectionData object, remove it from the list
                inputNames = inputNames(2:end);
                % get rid of inputs that might not have been assigned in data.settings
                % get length of inputs written in data.settings, trim the list in 'inputNames'
                inputNames = inputNames(1:numel( fieldnames( inputs ) ) );
                % order fields in inputs in case they are out of order
                inputs = orderfields(inputs, inputNames);
                % convert 'inputs to cell array'
                inputs = struct2cell(inputs);
                % run the function
                funcHandle(data, inputs{:});
            else
                % TODO - add support for anonymous functions? (through try-catch)
                taskLog = sprintf('%s%s\n', taskLog, 'Task not found in current path');
            end

        end
        
        % get how long the task took
        timeTaken = toc(singleTaskStopwatch);
        % write times taken into data.misc
        data.misc.taskTime(tasksToRun(k)) = timeTaken;
        % update log
        taskLog = sprintf('%s%s %.2f %s\n\n', taskLog, 'Task completed successfuly in', timeTaken, 'seconds');
        
    catch err
        taskLog = sprintf('%s%s\n', taskLog, 'Unsuccessful in running task');
        taskLog = sprintf('%s%s\n', taskLog, getReport(err, 'extended', 'hyperlinks', 'off'));
        
        % if no output argument was specified, display the report as a warning
        if nargout == 0
            warning(taskLog)
        end
    end
end


end


function saveProjections(data)
    % save projections and report
    
    % TODO - save things I want to put into the report into data.misc
    
    % TODO - allow saving of preview, but add a prefix?
    [path, name, ~] = saveResult(data, data.settings, 'proj');
    
    report = taskReport(data, data.settings);
    
    fileID = fopen([path, name, '_report', '.txt'], 'w+');
    fprintf(fileID, '%s', report);
    fclose(fileID);
end


