function [CoR] = corEstimManual(data)
%COROFFSETMANUAL allows user to manually set CoR of offset-sample datasets
%	When run, the function creates a simple GUI, which the user can control
%	using arrow keys and the return key (text instructions are displayed in
%	the GUI). The interface displays a reconstruction of the central slice
%	of the supplied dataset, given a particular CoR estimate. the user can
%	change this estimate and visually evaluate the resulting
%	reconstruction.
% Inputs:
%   data - projectionData object
% Outputs:
%   CoR - estimate of the location of the center of rotation within
%       projections. this estimate is also written into the GEOM
%       property of the projectionData object, so this output is mainly
%       for debugging purposes.
% Example:
%   CoR = corOffsetManual(data);
%   corOffsetManual(data);

% prior CoR estimate (using projection overlap)
if data.geom.offsetFlag
    current_CoR = corOffsetProj(data, 5);
else
    current_CoR = data.geom.CoR;
end

% extract sinogram from data (only middle slice)
sinogram = data.getSino;

% conversion to absorption values
trans2absorp(sinogram);

% create figure and axes to draw reconstruction in
ReconWindow = createFigure;
% display first reconstruction result
reconCurrentEstimate(sinogram, current_CoR, ReconWindow.Children(2))

% write CoR estimate into figure's appdata
setappdata(ReconWindow, 'current_CoR', current_CoR);
% write CoR estimate into figure's appdata
setappdata(ReconWindow, 'sinogram', sinogram);

go = true;
% save go to appdata so the figure can communicate when to stop while loop
setappdata(ReconWindow, 'go', go); 
while go
    if isvalid(ReconWindow) % if the figure hasn't been closed yet
        % get information on whether to continue
        go = getappdata(ReconWindow, 'go');
        % get current CoR estimate
        current_CoR = getappdata(ReconWindow, 'current_CoR');
        % update figure
        drawnow
    else
        go = false;
    end
end
% close figure (if it hasn't been closed already)
if isvalid(ReconWindow)
    close(ReconWindow)
end
% set current CoR estimate as output
CoR = current_CoR;

% write new CoR and reconstruction width values into data object
corInput(data, CoR);

end

function ReconWindow = createFigure
% get screen resolution
set(0,'units','pixels');
screenSize = get(0,'screensize');
% get the smaller of both dimensions
smaller = min(screenSize(3:4));

% create figure
ReconWindow = figure('Name', 'Manual estimation of the center of rotation', ...
                     'WindowStyle', 'normal', ...
                     'Units', 'pixels', ...
                     'Position', [smaller/4, smaller/4, smaller/2, smaller/2], ...
                     'MenuBar', 'none', ...
                     'Toolbar', 'figure', ...
                     'KeypressFcn', @keyPressed, ...
                     'DeleteFcn', @onClose, ...
                     'Tag', 'ReconWindow');
% create axes to draw into
ReconAxes = axes('Parent', ReconWindow, ... 
     'Title', 'Projection data', ...
     'Units', 'normalized', ...
     'Position',[0.09, 0.11, 0.8, 0.8], ...
     'Tag', 'ReconAxes');
% edit toolbar - somewhat convoluted a brittle, but no better solution yet
[major, minor, ~] = mcrversion;
if (major == 9 && minor >= 5) || major > 9 % in matlab R2018b and later
    % add axis interaction buttons to toolbar
    addToolbarExplorationButtons(ReconWindow)
    % remove axis toolbar
    axtoolbar(ReconAxes, 'default', ...
              'Visible', 'off', ...
              'Tag', 'ReconAxesToolbar');
end
% get all graphics objects in the figure
allHandles = findall(ReconWindow);
% find some menu items by their tooltip string, set visible to off
a = findall(allHandles,'ToolTipString','New Figure');
set(a,'Visible','Off')
a = findall(allHandles,'ToolTipString','Open File');
set(a,'Visible','Off')
a = findall(allHandles,'ToolTipString','Print Figure');
set(a,'Visible','Off')
a = findall(allHandles,'ToolTipString','Link Plot');
set(a,'Visible','Off')
a = findall(allHandles,'ToolTipString','Rotate 3D');
set(a,'Visible','Off')
a = findall(allHandles,'ToolTipString','Insert Legend');
set(a,'Visible','Off')
a = findall(allHandles,'ToolTipString','Edit Plot');
set(a,'Visible','Off')
a = findall(allHandles,'ToolTipString','Open Property Inspector');
set(a,'Visible','Off')

% create text w/ instructions
info = ['Use right arrow to increase CoR estimate by 1, ' ...
        'left arrow to decrease it by 1, ' ...
        'up arrow to increase it by 5, ' ...
        'or down arrow to decrease it by 5. ', ...
        'w and s shift the CoR by +-0.1 (experimental). ', ...
        'Use enter to confirm the current CoR estimate. ', ...
        'Closing using the red X also saves the currently displayed estimate'];
uicontrol('Parent', ReconWindow, ...
          'Style', 'text', ...
          'String', info, ...
          'Units', 'normalized', ...
          'Position',[0.05, 0.01, 0.90, 0.10], ...
          'Tag', 'TextInfo');
% TODO - make this prettier
end

function keyPressed(hObject, KeyData)
% get current CoR estimate
current_CoR = getappdata(hObject, 'current_CoR');

switch KeyData.Key
    case 'rightarrow'
        current_CoR = current_CoR + 1;
    case 'leftarrow'
        current_CoR = current_CoR - 1;
    case 'uparrow'
        current_CoR = current_CoR + 5;
    case 'downarrow'
        current_CoR = current_CoR - 5;
    case 'w'
        current_CoR = current_CoR + 0.1;
    case 's'
        current_CoR = current_CoR - 0.1;
    case 'return'
        % stop while loop in main function
        setappdata(hObject, 'go', false);
end
% save back to UserData
setappdata(hObject, 'current_CoR', current_CoR);

% get sinogram, proceed w/ reconstruction using current CoR estimate
sinogram = getappdata(hObject, 'sinogram');
reconCurrentEstimate(sinogram, current_CoR, hObject.Children(2))
% TODO - hObject.Children(2) can cause trouble if I change the GUI

end

function onClose(hObject, ~)

% housekeeping
sinogram = getappdata(hObject, 'sinogram');
delete(sinogram) % delete data in handle object

delete(hObject)
astra_clear

end

function reconCurrentEstimate(sinogram, current_CoR, ReconAxes)

% make a copy of the sinogram
cur_sino = copy(sinogram);

% change CoR to current estimate
cur_sino.geom.edit('CoR', current_CoR)
shiftFromOrigMiddle = cur_sino.geom.CoR_shift; % shift of CoR from original projection's midpoint
if cur_sino.geom.offsetFlag
    % calculate width of extended detector
%     extendedWidth = round( (cur_sino.geom.width - (cur_sino.geom.CoR - 0.5) ) * 2);
    extendedWidth = round( (cur_sino.geom.width - cur_sino.geom.CoR) * 2); % always even
    % write into data.geom
    cur_sino.geom.edit('reconWidth', extendedWidth);

    % if offset scan, transform projections, get new geometry
    virtualDetTransf(cur_sino, true);
    % if offset scan, apply weights to data and pad them so CoR is in the middle
    offsetWeighPad(cur_sino, true);
else
    shiftProjs(cur_sino); % shift to compensate for CoR shift
end

% % pad to get rid of truncation artifacts
% truncationPadding(cur_sino, 10);

cur_sino.settings.addRec('ASTRA', 'FDK', ...
         'filterType', 'cosine', 'freqCutoff', 0.85, 'limitGPU', true);
reconstruct_ASTRA_FDK(cur_sino);

% % show detail of middle
% mid = cur_sino.geom.reconWidth/2;
% imshow(cur_sino.tomo( (mid-80):(mid+80), (mid-80):(mid+80)), [], 'Parent', ReconAxes);

% get current zoom
if ~isempty(ReconAxes.Children)
    lims = get(ReconAxes,{'xlim','ylim'});
else
    lims = {[0.5 size(cur_sino.tomo, 1)+ 0.5], [0.5 size(cur_sino.tomo, 2)+ 0.5]};
end
    
% show reconstruction
imshow(cur_sino.tomo, [], 'Parent', ReconAxes);

% update axes title etc.
% disable axes rulers
ReconAxes.XAxis.Visible = 'off';
ReconAxes.YAxis.Visible = 'off';
% set title of axes
title = sprintf('%s: %.1f, %s: %.1f', 'Current CoR estimate', current_CoR, ...
                                  'Shift from middle', shiftFromOrigMiddle);

ReconAxes.Title.String = title;

% set axes to previous zoom
zoom reset
set(ReconAxes, {'xlim','ylim'}, lims)

% housekeeping
delete( cur_sino )

end % TODO - fix error in case i have offset data and go beyond half

% TODO - make the code clearer