function [CoR] = corEstimRec(data, slicesToAverage, broadSearch, overwriteFlag)
%COROFFSETREC estimates position of rotation center in offset-sample data
%   This function takes as input a projectionData object, extracts the
%   sinogram of the middle slice of data saved within it, and evaluates
%   the reconstruction of this slice at different CoR estimates. More
%   precisely, it takes the average magnitude of the reconstruction's
%   gradient as an objective metric, detrends the series of these averages
%   using a simple moving average filter, and then finds a maximum in this
%   detrended series. This maximum should correspond to the location of the
%   center of rotation.
%   NB: currently, ASTRA toolbox's FDK implementation is used to
%   reconstruct the data. The filter used is ram-lak with a 0.5 frequenc
%   cutoff currently.
%   NB: This function assumes the CoR to be in the left part of projections
% Inputs:
%   data            - projectionData object
%   slicesToAverage - number of slices around the middle slice to average
%   into one sinogram. This may help reduce the contribution of noise to
%   the CoR estimation. (positive scalar)
%   broadSearch     - logical flag. If false, the function COROFFSETPROJ is
%   used for a prior rough estimate, and COROFFSETREC is used only to
%   refine this rough estimate by searching for the optimum in a small area
%   around it. If true, the function searches in a much wider area based on
%   the information in scan metadata.
% Outputs:
%   CoR - estimate of the location of the CoR (as the index of a
%   detector bin). This CoR estimate is also written into the GEOM 
%   property of data (scalar)
% Example:
%   CoR = corOffsetRec(data, 1, false);
%   corOffsetRec(data, 1, false);
% 
% NB:
% This function assumes an offset of the sample (not the detector), and
% that the sample (and CoR) has been offset to the left

% TODO - check inputs
% TODO - if slicestoaverage is not specified, set default
% etc.

if nargin < 4
    overwriteFlag = false;
    if nargin < 3 % make broadSearch false by default
        broadSearch = false;
    end
end

% determine range of sinograms around middle slice to be averaged
middle = (data.geom.height + 1) / 2;
sinoRange = round( [middle - (slicesToAverage-1)/2, middle + (slicesToAverage-1)/2]);
% get (averaged) sinogram from data
sinogram = data.getSino(sinoRange, 1); % another projectionData object
% convert to absorption values
trans2absorp(sinogram)

% save geom and params to separate variables to make code more readable
geom = sinogram.geom;
params = sinogram.params;

factor = 8 / params.binning; % factor to downsample reconstructions by

if broadSearch % search for CoR in a wide region - time-consuming
    if isfield(params, 'delta_T')
        delta_T = params.delta_T;
    else
        delta_T = 0;
    end
    est_CoR = round((geom.width+1)/2 - delta_T*1000/geom.pix_size ); % estimation of CoR location based on info about offset in metadata
    buffer = round( 160 / sinogram.params.binning);
else % search for CoR in a small area - relies on prior estimate
    est_CoR = corOffsetProj(data, 5); % rough CoR estimate from projection overlap
    buffer = 10;
end

if sinogram.geom.offsetFlag % there has to be some minimum overlap
    % range(1) = max( est_CoR - buffer, round(geom.width*0.01) );
%     range(1) = max( est_CoR - buffer, 10 );
%     range(1) = max( est_CoR - buffer, 5 );
    range(1) = max( est_CoR - buffer, 2 );
    range(2) = min( est_CoR + buffer, floor(sinogram.geom.width/2) ); % assumption - CoR is in the left part of the sinogram
else % if not offset data
    range(1) = est_CoR - buffer;
    range(2) = est_CoR + buffer;
end

% pre-filter with a gaussian kernel to suppress noise
sinogram.proj = imgaussfilt( sinogram.proj, 3 );

% % Histogram equalization
% sinogram.proj = histeq(sinogram.proj);
% sinogram.proj = permute( adapthisteq(permute(sinogram.proj, [3,2,1])), [3,2,1]);

% Proceed with CoR estimation
step = 1; % increment of the search
guess_range = range(1):step:range(2); % TODO: investigate these values for the CoR

CoR = goThroughRecs(sinogram, guess_range, factor, false);

% step = 0.1; % increment of the search
% fine_range = 3;
% guess_range = (CoR - fine_range):step:(CoR + fine_range); % TODO: investigate these values for the CoR
% 
% CoR = goThroughRecs(sinogram, guess_range, factor, true);

if overwriteFlag
    corInput(data, CoR);
end

end


function [CoR] = goThroughRecs(sinogram, guess_range, factor, fix_rec_size)

% M = zeros(1, numel(guess_range)); % preallocate
Qin = zeros(1, numel(guess_range));

parfor (k = 1:length(guess_range), 8) % limit number of workers to prevent crashes
% for k = 1:length(guess_range) % range of looking for the CoR
    
    cur_sino = copy(sinogram);
    cur_sino.geom.edit('CoR', guess_range(k))
    
    if cur_sino.geom.offsetFlag % change reconstruction width along with it
%         extendedWidth = round( (cur_sino.geom.width - (cur_sino.geom.CoR - 0.5) ) * 2);
        if fix_rec_size
            extendedReconWidth = round( (cur_sino.geom.width - guess_range(end)) * 2) - 20; % always even
        else
            extendedReconWidth = round( (cur_sino.geom.width - cur_sino.geom.CoR) * 2); % always even
        end
        cur_sino.geom.edit('reconWidth', extendedReconWidth);

        virtualDetTransf(cur_sino, true);
        offsetWeighPad(cur_sino, true);
        truncationPadding(cur_sino, 10);
%     else
%         shiftProjs(cur_sino); % TODO - use postalignment
    end
    
    
%     proj_geom = astra_create_proj_geom('fanflat', det_width, det_count, angles, source_origin, origin_det);
    proj_geom = astra_create_proj_geom('fanflat', ...
                    cur_sino.geom.cb_mag / factor, ... % detector width
                    cur_sino.geom.width, ... % detector count
                    cur_sino.geom.angles, ... % angles
                    cur_sino.geom.SOD/ (cur_sino.geom.cb_mag*factor), ... % source-origin distance
                    cur_sino.geom.ODD/ (cur_sino.geom.cb_mag*factor)); % origin-detector distance
    proj_geom = astra_geom_postalignment(proj_geom, cur_sino.geom.CoR_shift);
    
    volSize = round( cur_sino.geom.reconWidth/factor );
    vol_geom = astra_create_vol_geom(volSize, volSize);
    
    temp_sino = cur_sino.proj;
    temp_sino = permute(temp_sino, [3, 2, 1]); % transpose for ASTRA
    
    sino_id = astra_mex_data2d('create', '-sino', proj_geom, temp_sino);
    rec_id = astra_mex_data2d('create', '-vol', vol_geom);

    cfg = astra_struct('FBP_CUDA');
    cfg.ProjectionDataId = sino_id;
    cfg.ReconstructionDataId = rec_id;
    cfg.FilterType = 'ram-lak'; % Options: 'ram-lak' (default), 'shepp-logan', 'cosine', 'hamming', 'hann', 'none', 'tukey', 'lanczos', 'triangular', 'gaussian', 'barlett-hann', 'blackman', 'nuttall', 'blackman-harris', 'blackman-nuttall', ?eflat-top?f, 'kaiser', 'parzen', 'projection', 'sinogram', 'rprojection', 'rsinogram'.
    cfg.FilterD = 1;
%     cfg.FilterD = 0.5; % TODO - test different filters
    
    fbp_id = astra_mex_algorithm('create', cfg);
    astra_mex_algorithm('run', fbp_id);
    tomo = astra_mex_data2d('get', rec_id);
    
    astra_mex_data2d('delete', sino_id, rec_id);
    astra_mex_algorithm('delete', fbp_id);
    
%     [Gmag, ~] = imgradient(tomo, 'central'); % TODO - use gpu to calculate gradient
    
    width = size(tomo, 1); % remove areas outside FoV
    [x ,y] = meshgrid(1:width, 1:width); 
    mask = sqrt((x-(width+1)/2).^2 + (y-(width+1)/2).^2) < (width/2); % create 2D circular mask
    
    % save current values of the metrics
%     M(k) = mean( Gmag(mask) );
%     Qia(k) = sum( abs(tomo(:)) ) / sum(mask(:));
    Qin(k) = sum( -tomo(tomo < 0) ) / sum(mask(:));
%     Qh(k) = entropy(tomo(mask));
    
%     % Visualize - needs for
%     subplot(2, 2, 1); imshow(tomo, []); title('Reconstruction')
%     subplot(2, 2, 2); imshow(Gmag, []); title('Gradient')
%     subplot(2, 1, 2); plot(guess_range(1:k), M(1:k)); title('Gradient averages')
%     xlabel('CoR estimate'); ylabel('Average gradient')
%     pause
end

% Mtrend = savitzkyGolayFilt(M, 3, 0, 11); % (input_data, degree, differentiation_order (0 means smoothing), frame_size)
% Mdetrend = M - trend;
% Qtrend = savitzkyGolayFilt(Qin, 3, 0, 11); % (input_data, degree, differentiation_order (0 means smoothing), frame_size)
% Qdetrend = Qin - Qtrend;

[~, maxIndex] = min( Qin );

% get the CoR corresponding to this maximum
CoR = guess_range(maxIndex+1);

% % Visualization
% figure;
% subplot(2,1,1);
% plot(guess_range, M./max(M)); title('Sum gradient magnitude of reconstruction')
% hold on
% plot(guess_range, trend./max(M));
% xlabel('AoR position [-]'); ylabel('Sum of gradient magnitude relative to maximum')
% hold on; plot(guess_range(maxIndex), M(maxIndex)./max(M), '*r')
% legend('Average gradient magnitude', ['Estimated AoR at index ' num2str(CoR)])
% xlim(range)
% subplot(2,1,2);
% plot(guess_range, detrend./max(detrend)); title('Sum gradient magnitude of reconstruction (de-trended)')
% hold on
% plot(guess_range(maxIndex), detrend(maxIndex)./max(detrend), '*r')
% hold off
% xlabel('AoR position [-]'); ylabel('Detrended magnitude average relative to maximum')
% legend('Detrended average magnitude signal', ['Estimated CoR at index ' num2str(CoR)])
% xlim(range)

end









