function [CoR] = corOffsetProj(data, numEstimates)
%COROFFSETPROJ performs rough estimate of CoR in offset sample data
%   This is a utility function used as a rough prior estimate in more
%   precise methods, such as the function COROFFSETREC. It is based on the
%   similarity of the overlap of two projections 180 apart.
%   NB: This function assumes the CoR to be in the left part of projections
% Inputs:
%   data         - projectionData object
%   numEstimates - number of projection pairs used for estimation of the
%       CoR (if this is e.g. 3, then the 3 pairs used for estimation are
%       0-180, 61-241, and 120-300)
% Outputs:
%   CoR - the rough estimate of the location of the CoR (as the index of a
%       detector bin). Unlike in other functions, this CoR estimate is NOT
%       also written into the GEOM property of data
% Example:
%   CoR = corOffsetProj(data, 5)

% NB: assuming CoR is on the left
% NB: very rough estimation, so I'm not transforming projections
% NB: might fail cos of cone beam effects sometimes (or for some samples)

% NB: - assumes some min. overlap, assumes equispaced angular sampling of
% projections...

% TODO - validate numEstimates - whole, positive, less than num_proj/2...
% TODO - make a default value of numEstimates

% TODO - only a utility function of corOffsetRec and corOffsetManual


geom = data.geom;
params = data.params;

% choose indexes of projections to compare
projs_to_compare = round( linspace(1, geom.num_proj/2, numEstimates+1));
projs_to_compare(end) = [];

% index of projection at 180 degrees (opposite 1st projection)
[~, opposite_proj] = min(abs(geom.angles - (geom.angles(1)+pi)));

% estimation of CoR location based on scan metadata
est_CoR = round( (geom.width+1)/2 - params.delta_T * 1000 / geom.pix_size );

buffer = 120 / params.binning; % determines area around est_CoR, where I will check for the CoR
% 120 is highly composite, so it should be whole after division by most binning values
% TODO - make the minimum of range(1) an input or dynamic based on binning?
range(1) = max( est_CoR - buffer, floor(data.geom.width/20) ); % there has to be some overlap, so here I am saying the overlap should be at least 30 pixels
% TODO - make minimum depend on binning or something?
range(2) = min( est_CoR + buffer, floor(geom.width/2) ); % I am assuming the CoR is in the left part of the sinogram
% these are indexes of detector bins (indexing starts from 1)

% Proceed with CoR estimation
step = 1; % increment of my search
guess_range = range(1):step:range(2); % I will investigate these values for the CoR

% % to speed up the process, trim projections of larger datasets
% trimRange = 256; % will be trimmed to +- trimRange around the middle
% middle = round( (geom.height + 1) / 2 );
% vertRange = max(1, (middle - trimRange)):min(geom.height, (middle + trimRange));

% preallocate vector of CoR estimates
cor_estimates = zeros(1, numEstimates);
for k = 1:numEstimates
    % 1st projection, take only overlapping part, also trim vertically
    proj1 = data.proj(:, 1:(2*range(2)), projs_to_compare(k));
    % opposite projection, take only overlap part, also trim vertically
    proj2 = data.proj(:, 1:(2*range(2)), ...
                      projs_to_compare(k) - 1 + opposite_proj);

    % flip the opposite projection
    proj2 = fliplr( proj2 );
    
    % filter projections by an std filter
    proj1 = stdfilt( proj1 );
    proj2 = stdfilt( proj2 );
    % TODO - try LoG
    
    D = zeros(1, length(guess_range));
    parfor m = 1:length(guess_range)
        % get only the overlapping part
        overlap1 = proj1(:, 1:(2*guess_range(m)));
        overlap2 = proj2(:, (end-(2*guess_range(m))+1):end);
        
%        % Visualize - needs for to work
%        imshow(abs(overlap1 - overlap2), []); title(['CoR: ' num2str(guess_range(m))])
%        pause

       % Different metrics:
       % Mean squared error
%        difference = overlap1 - overlap2;
%        D(m) = sum( difference(:).^2 ) / length(difference(:));
       % Standard deviation of difference
       difference = overlap1 - overlap2;
       D(m) = std(difference(:));
       % Cosine distance
%        D(m) = 1 - (overlap1(:)' * overlap2(:)) / (norm(overlap1(:)) * norm(overlap2(:)));
       % calculate Pearson correlation coefficient (invert to look for min)
%        covariance = cov(overlap1, overlap2);
%        D(m) = 1 - covariance(1,2) / (std(overlap1(:))*std(overlap2(:)));
    end
    
    
%     % Visualize
%     plot(guess_range, D); title(['difference metric for overlap at different CoR values ' num2str(k)])
%     pause
    
    % find the minimum in the D vector
    [~, index] = min(D);
    % the CoR corresponding to the index of the minimum will be my guess
    if isempty(index)
        cor_estimates(k) = 1;
    else
        cor_estimates(k) = guess_range(index(1));
    end
end

% take the median estimate
CoR = median(cor_estimates);

% TODO - make this more reliable. Currently, this is very crude

end