function [CoR] = corOffsetSino(data, slicesToAverage, overwriteFlag)
%COROFFSETSINO estimates offset-sample CoR through sinogram symmetry prop.
%   This function takes as input a projectionData object, extracts a
%   sinogram of the middle slice of data saved in it, and estimates the
%   center of rotation of the sample based on the symmetric property of
%   sinograms around the center of rotation (in short, both halves of a
%   360 sinogram should contain the same information. For offset data,
%   this holds true only for a small region around the CoR, where the
%   sample is scanned twice).
%   Inputs:
%       data - projectionData object
%       slicesToAverage - scalar, denoting how many sinograms around the
%       middle slice should be averaged. This may help reduce the
%       contribution of noise to CoR estimation in some cases
%   Outputs:
%       CoR - estimate of the location of the CoR (as the index of a
%       detector bin). This CoR estimate is also written into the GEOM 
%       property of data (scalar)
%   Example:
%       CoR = corOffsetSino(data, 1);
%       corOffsetSino(data, 1);
% 
% NB:
% This function assumes an offset of the sample (not the detector), and
% that the sample (and CoR) has been offset to the left
% 
% Based on 'Calibration Method of Center of Rotation Under the Displaced 
% Detector Scanning for Industrial CT'
% by Qiang Lin et al (2018)
% DOI: https://doi.org/10.1016/j.nima.2018.11.131


% TODO - check inputs

if nargin < 3
    overwriteFlag = false;
end

% determine range of sinograms around middle slice to be averaged
middle = (data.geom.height + 1) / 2;
sinoRange = round( [middle - (slicesToAverage-1)/2, middle + (slicesToAverage-1)/2]);
% get (averaged) sinogram from data
sinogram = data.getSino(sinoRange, 1); % another projectionData object

% trans2absorp(sinogram) % TODO - seems unnecessary here

% save geom and params to separate variables to make code more readable
geom = sinogram.geom;
params = sinogram.params;
% extract the sinogram from the projectionData object, turn it the right way
sino = permute(sinogram.proj, [3 2 1]);


% estimation of CoR location based on based info about offset in metadata
est_CoR = round( (geom.width+1)/2 - params.delta_T * 1000 / geom.pix_size );

buffer = round(80 / sinogram.params.binning);

% there has to be some minimum overlap
% range(1) = max( est_CoR - buffer, floor(sinogram.geom.width/20) );
% range(1) = max( est_CoR - buffer, 10 );
range(1) = max( est_CoR - buffer, 2 );

% I am assuming the CoR is in the left part of the sinogram
range(2) = min( est_CoR + buffer, floor(sinogram.geom.width/2) );
% these are indexes of detector bins (indexing starts from 1)


% Proceed with CoR estimation
step = 1; % increment of my search
guess_range = range(1):step:range(2); % I will investigate these values for the CoR

D = zeros(1, length(guess_range)); % preallocate D
parfor k = 1:length(guess_range)
% for k = 1:length(guess_range)
    % TODO - change loop variable to k
    
    % CoR guess is indexing from 1, overlap is ( (CoR_guess)*2 - 1 ) pixels
    % I need to transform projections to a virtual detector that is perpendicular to the ray passing through the center of rotation
    P_left = sino(:, 1:floor(guess_range(k))); % get the left part of the sinogram just like that (because the virtual detector is closer to the source here)
    % for the right part, I need to find the corresponding points. I will
    % get positions on the real detector, so I can index the sinogram directly
    theta1 = atan( ((geom.width+1)/2 - guess_range(k)) / (geom.SOD + geom.ODD)); % the angle between the virtual and real detector
    theta2 = atan( (geom.SOD + geom.ODD) ./ (geom.width/2 - (1:guess_range(k))) ); % angles at which rays hit the real detector at the CoR
    theta3 = pi - theta1 - theta2; % pi minus the angle at which rays hit the virtual detector on the left side of CoR
    new_FDD = (geom.SOD + geom.ODD) / cos(theta1); % distance from source to the detector bin at the center of rotation estimate
    left_det = ( guess_range(k) - (1:floor(guess_range(k))) ) .* ( sin(theta2) / sin(theta3) ); % distances of detector bins on the virtual detector from the CoR
    theta6 = atan(new_FDD ./ left_det); % angle at which rays hit the virtual detector on the right side of CoR
    theta5 = pi - theta6 - theta1; % pi minus the angle at which rays hit the real detector on the right side of CoR
    right_det = floor(guess_range(k)) + ( guess_range(k) - (floor(guess_range(k)):-1:1) ) .* ...
                (sin(theta2) ./ sin(theta3)) .* ( sin(theta6) ./ sin(theta5)); 
    % indexes of query points at the real detector on the right side of CoR
    
    % Get the sinogram values at the calculated query points
    [X, Y] = meshgrid(1:geom.width, 1:geom.num_proj);
    P_right = interp2(X, Y, sino, repmat( right_det, geom.num_proj, 1), Y(:, 1:length(right_det) ) , 'linear');   
    
    % get sums of left and right side
    sumP_left = sum(P_left, 1);
    sumP_right = fliplr(sum(P_right, 1)); % careful, flipped
    
    % Different options of measuring similarity (std & cosine seem best):
    
%     % calculate MSE between left and right side
    D(k) = sum( (sumP_left - sumP_right).^2 ) / length(sumP_left);
    % calculate euclidean distance (L2 norm), divide by number of samples
%     D(k) = sqrt( sum( (sumP_left - sumP_right).^2 ) ) / length(sumP_left); % division by N
%     % calculate standard deviation of difference between left and right
%     Dstd(k) = std(sumP_left - sumP_right);
%     % calculate cosine similarity (invert to look for min - cosine distance)
%     Dcos(k) = 1 - (sumP_left * sumP_right') / (norm(sumP_left) * norm(sumP_right));
%     % calculate Pearson correlation coefficient (invert to look for min)
%     covariance = cov(sumP_left, sumP_right);
%     Dcov(k) = 1 - covariance(1,2) / (std(sumP_left)*std(sumP_right));
    

%     % Visualization - needs for, doesn't run in parfor
%     subplot(2,1,1)
%     plot(sumP_left);
%     hold on;
%     plot(sumP_right);
%     [~, ind] = min(D(1:k));
%     
%     hold off
%     title(['Current CoR estimate = ' num2str(guess_range(k))])
%     subplot(2, 1, 2)
%     plot(guess_range(1:k), D(1:k))
%     hold on
%     plot(guess_range(ind), D(ind), '*r')
%     title(['Minimum of s - D curve: ' num2str(D(ind)) ' at index ' num2str(guess_range(ind)) ' (marked in red)'])
%     hold off
%     
% %     imshow([P_left circshift(P_right, round(length(geom.angles)/2), 1)],[])
% %     title(['CoR: ' num2str(CoR_guess)])
%     pause
end
[~, ind] = min(D); % get the minimum of the function

% get the CoR corresponding to the minimum
CoR = guess_range(ind);

% % Visualization
% figure;
% plot(guess_range, D)
% hold on
% plot(guess_range(ind), D(ind), '*r')
% title(['Minimum of difference curve at index ' num2str(guess_range(ind)) ' (marked in red)'])
% xlabel('Index of the center of rotation from projection left [pixels]');
% ylabel('L2 norm of sinogram halves') % TODO - change label if I change the way D is calculated
% hold off

if overwriteFlag
    corInput(data, CoR);
end

% for debugging only !!!!
assignin('base', 'D', D);
assignin('base', 'sinoRange', guess_range);

end

