function offsetWeighPad(data, padFlag)
%OFFSETWEIGHPAD weighs and pads offset data to prepare them for FDK recon.
%   The function takes as input a 2D sinogram or 3D array of projections
%   with an offset center of rotation, as well as information about this
%   offset, and properly pre-processes them so that they are suitable for
%   reconstruction via standard FDK. Pre-processing includes weighing the
%   overlapping region around the CoR, as well as padding the shorter side 
%   of the projections with 0s (if specified through padflag)
%   NB: this function assumes CoR is in the left part of projections
%   Inputs:
%       data - projectionData object
%       padFlag - logical flag denoting whether to pad projections or not
%   Example:
%       offsetWeighPad(data, 1);

% TODO - check inputs (especially padflag)

% --- Code ---
geom = copy(data.geom); % extract geom from projectionData object;

% Weigh data
w_range = round(2*geom.CoR);
weight = 0.5 .* sin( pi .* linspace(-geom.CoR, geom.CoR, w_range) / w_range) + 0.5; % TODO - check

if ndims(data.proj) == 3 % if proj is a 3D array of projections
    weight = repmat(weight, geom.height, 1, geom.num_proj); % repeat weight vector for the whole array
else % else it is a 2D sinogram
    weight = repmat(weight, geom.num_proj, 1); % repeat weight vector for the whole array
end
% apply weights to data
data.proj(:, 1:size(weight,2), :) = data.proj(:, 1:size(weight,2), :) .* weight;
clear weight % save memory

% get width of extended detector (calculated along with CoR estimate in CoR estimation functions)
extendedWidth = geom.reconWidth;

if padFlag % Pad data
    if ndims(data.proj) == 3 % if proj is a 3D array of projections
        pom = zeros(geom.height, (extendedWidth - geom.width), geom.num_proj, 'single');
        data.proj = [pom, data.proj]; % TODO - is there a more efficient way to concatenate arrays?
    else % else it is a 2D sinogram
        pom = zeros(geom.num_proj, (extendedWidth - geom.width), 'single');
        data.proj = [pom, data.proj];
    end
end

% update geometry - projections are now wider (if padded)
data.geom.edit('width', size(data.proj, 2));
% if projections were padded, update info about CoR
if padFlag
    data.geom.edit('CoR', data.geom.CoR + extendedWidth - geom.width);
end

end

