function [transf] = virtualDetTransf(data, keepVoxelSize)
%VIRTUALDETTRANSF transforms offset-sample projections to offset-detector
%   the function takes offset-sample projection data and transforms it into
%   an offset-detector geometry. The utility of this function is for the
%   projections to be compatible with a regular FDK-type reconstruction
%   scheme. When the sample is offset, the central ray is not orthogonal to
%   the detector. This function transforms projections onto a virtual
%   detector, which is slightly tilted when compared to the real detector,
%   and is orthogonal to the central ray.
%   Inputs:
%       data          - object containing the projection data, a
%                       scanGeometry object, etc.
%       keepPixelSize - scalar boolean. If true, transformed projections
%                       will have the same pixel size as original ones
%   Outputs:
%       transf - structure containing information on the transformation of
%           projections onto a virtual detector. Contains the fields:
%           transf.known_x - horizontal coordinates of known pixels
%               (projected onto the virtual detector, not equispaced)
%           transf.query_x - horizontal coordinates of query points on
%               the virtual detector (equispaced)
%           transf.known_y - vertical coordinates of known pixels
%               (projected onto the virtual detector, not equispaced)
%           transf.query_y - vertical coordinates of query points on
%               the virtual detector (equispaced)
%           Information in the transf structure is used inside the
%           VIRTUALDETTRANSF function as inputs for interp2 as such:
%               proj_tr(:, :, k) = interp2(k_points_x, k_points_y, proj(:, :, k), q_points_x, q_points_y, 'method', ...);
%               where proj(:, :, k) is the k-th projection in a 3D array
%           this output argument is optional, only for debugging
%   Example:
%       transf = virtualDetTransf2D(data);

% --- Check inputs
% TODO - do this section

if nargin < 2
    keepVoxelSize = true;
end


% --- Code ---
geom = data.geom; % reference geometry field by geom (to make code shorter)

% Calculate parameters of virtual detector the projection will be mapped to (all assuming 1 pixel as the unit)
SDD = (geom.SOD + geom.ODD) / geom.pix_size; % total distance from source to detector, in pixels
det_tilt = atan( - geom.CoR_shift / SDD ); % tilt of new detector relative to the old one

% calculate SDD for the new detector - pythagoras' theorem
new_SDD = sqrt( SDD^2 + geom.CoR_shift^2 ); 

% calculating angles at detector bins of the old detector
pixIndRelativeToCoR = (1:geom.width) - geom.CoR; % (signed) distances of old detector bins from new CoR

pixelAngles = atan(pixIndRelativeToCoR ./ SDD); % angle of ray at each detector bin relative to middle ray

% use new_SDD, CoR shift, and old_bin_angles to find locations corresponding to old detector bins on the new detector
new_det_points = pixIndRelativeToCoR .* ...
                 sin(pi/2 - det_tilt + pixelAngles) ./ ...
                 ( sin(pi/2 - pixelAngles) ); % where the bins of the old detector project on the new one (sine rule)
% calculate coordinates of the calculated points
changeDistFromSource = sin(det_tilt) .* new_det_points; % change in distance of detector columns from source

% % visualize the shift (from above)
% new_hor_dist = cos(det_tilt) .* new_det_points; % horizontal coordinates of old detector bin projections on the virtual detector
% figure;
% hold on
% plot(0, 0, 'bx'); % center of rotation projection on detector
% plot(pixIndRelativeToCoR, zeros(1, length(pixIndRelativeToCoR)), 'g'); % real detector
% plot(new_hor_dist, changeDistFromSource, 'r'); % new virtual detector
% plot([-geom.CoR_shift, 0], [-SDD 0], '--k'); % line from source to CoR
% plot(-geom.CoR_shift, -SDD, 'kx'); % source point
% plot([pixIndRelativeToCoR(1), -geom.CoR_shift, new_hor_dist(end)], [0, -SDD,  changeDistFromSource(end)], 'm') % left beam
% title('Real (Green) and Virtual (Red) Detector, Viewed from Above')
% xlabel('x coordinate [pixels]')
% ylabel('y coordinate [pixels]')
% legend('Center of Rotation', 'Real Detector', 'Virtual detector', 'Ray perpendicular to Virtual Detector', 'Source', 'X-ray Beam')
% hold off

% NB: new_det_points are points along an axis parallel to the virtual detector
% deltaSDD and new_hor_dist are coordinates of those points in the same coordinate system as the real detector

new_width = new_det_points(end) - new_det_points(1) + 1; % how wide is the transformed detector (keeping original numbers as the unit)
% NB: +1, because the middle of the 1st bin is 0.5 away from the edge, same with the last bin

% calculating known and query X coordinates
transf.known_x = repmat(new_det_points, geom.height, 1); % x coordinates of known points

if keepVoxelSize
    % make it so the voxel size will be the same as before transform
    newPixSize = geom.pix_size; % keep old pixel size
    query_x = linspace(new_det_points(1), ...
                       new_det_points(1) + geom.width - 1, geom.width);
else
    % use the whole width of the detector, but change the voxel size
    newPixSize = geom.pix_size .* (new_width / geom.width);  % new pixel size
    query_x = linspace(new_det_points(1), new_det_points(end), geom.width); % looking at equispaced locations
end

transf.query_x = repmat(query_x, geom.height, 1); % x coordinates of query points

% calculating known and query Y coordinates - these I have to do at the real detector (to have the point matrices in meshgrid format)
known_y = (-(geom.height-1)/2):((geom.height-1)/2); % known points - equispaced at real, but non-equispaced at virtual detector
transf.known_y = repmat (known_y', 1, geom.width);

% new_height =  geom.height * new_width/geom.width; % in order to have isotropic voxels, I should calculate the height again
% query_y = linspace(-(new_height-1)/2, (new_height-1)/2, geom.height); % equispaced query points at the virtual detector, not equispaced at the real one

query_y = linspace(-(geom.height-1)/2, (geom.height-1)/2, geom.height);
magnifications = (SDD + changeDistFromSource) ./ SDD; % difference in magnification at each bin when compared to central bin (still considering original source)
transf.query_y = repmat(query_y' , 1, geom.width) ./ repmat(magnifications, geom.height, 1);

% transform and interpolate projections.
proj = data.proj; % temporarily put it outside the object, so it can go to parfor
data.proj = [];
if geom.height == 1 % if it is only 1 sinogram
    parfor k = 1:geom.num_proj
        proj(:, :, k) = interp1(transf.known_x, ...
                               proj(:, :, k), ...
                               transf.query_x, ...
                               'spline', 'extrap');
    end
else % if interpolating a 3D array of projections (not only 1 sinogram)
    
    % value of pixels that are out of range of original projections
    % get values
    tempProj = proj(1, :, :);
    topMean = mean( tempProj(:) ); % get mean value of top sinogram
    tempProj = proj(end, :, :);
    botMean = mean( tempProj(:) ); % same for bottom
    
    tempProj = interp2(transf.known_x, transf.known_y, ... % get indices of extrapolated pixels at the top and bottom
                       zeros(geom.height, geom.width), ...
                       transf.query_x, transf.query_y, ...
                       'spline', 1); % make mock transform
    extrapIndices = find( tempProj ); % find linear indices of extrapolated values at the top
    [rows, ~] = ind2sub([geom.height, geom.width], extrapIndices); % get corresponding row indices
    extrapIndicesTop = extrapIndices(rows < geom.height/2); % extract indices on top of projection
    extrapIndicesBot = extrapIndices(rows > geom.height/2); % extract indices on bottom of projection
    
    parfor k = 1:geom.num_proj
        tempProj = proj(:, :, k);
        tempProj = interp2(transf.known_x, transf.known_y, tempProj, ...
                           transf.query_x, transf.query_y, 'spline', 0);
        tempProj(extrapIndicesTop) = topMean;
        tempProj(extrapIndicesBot) = botMean;
        
        proj(:, :, k) = tempProj;
    end
end
data.proj = proj;
% Calculation of new geometry values
extractTransfGeom(data, new_SDD, newPixSize, det_tilt, query_x);

end


function extractTransfGeom(data, new_SDD, newPixSize, det_tilt, query_x)
%extractTransfGeom is a function for calculating the geometry parameters of
%projections transformed using the virtualDetTransf2D function. It works in
% a similar way to that of extractGeom.
% TODO - update description

% --- Code ---
geom = data.geom;

transf_SOD = sqrt(geom.SOD^2 + ( ( abs(geom.CoR_shift) * geom.pix_size) ...
                  / geom.cb_mag )^2);
transf_ODD = new_SDD * geom.pix_size - transf_SOD; % new_SDD is in pixels, therefore i need to multiply it by the pixel size (the old one)
% cb_mag and vox_size are dependent properties - will change automatically

transf_angles = geom.angles + det_tilt; % add the difference between real and virtual detector to the information about angles

[~, ind] = min( abs( query_x ) );
val = query_x( ind );
transf_CoR = ind - val; % position of CoR on virtual detector
% since in new_det_edges, the location of the edges is calculated relative to the CoR, I can just take the 1st bin (which will be negative)
% and take its absolute value to get the distance of CoR from the detector' left edge (and add +1 to make it an index)

% reconWidth and reconHeight will stay the same

% write into geom
data.geom.edit(...
    'pix_size', newPixSize, ...
    'SOD', transf_SOD, ...
    'ODD', transf_ODD, ...
    'angles', transf_angles, ...
    'CoR', transf_CoR);

end


