function [ kernel ] = filterKernel(width, type, cutoff, apoOnly)
%FILTERKERNEL outputs a filter kernel for FBP-type reconstruction
%   This function creates a filter kernel for FBP-type reconstruction in
%   the frequency domain
%   Inputs:
%       width - scalar, the width of projection data to be filtered (NB:
%       the kernel on the output will have length 2^nextpow2(2*width))
%       type - character string describing the type of apodization low-pass
%       window to be used. Options include: 'cosine', 'hamming', 'hann',
%       'shepp-logan', 'ram-lak', and 'none'
%       cutoff - scalar larger than 0. Denotes the cutoff frequency as a
%       ratio of the nyquist frequency.
%       apoOnly - logical flag signifying whether to output only the low
%   Outputs:
%       kernel - vector, the frequency response of the desired filter
%   Example:
%       kernel = filterKernel(1000, 'cosine', 0.85, false);



%filterKernel outputs a frequency domain FBP-type ram-lak filter kernel to 
%be used in an FDK reconstruction. The output kernel has length
% 2^nextpow2(2*width). It is possible to also set a frequency cutoff and
% low-pass apodization window. Optionally, only the low-pass window can be
% output.
%   TODO - describe inputs and outputs

% TODO - check inputs

% TODO - check that everything works as intended
% TODO - add 'none' option

% --- Code ---
% no. of samples in kernel
% take next higher power of 2 of double the projection width
n = 2^nextpow2(2 * width);
% just numbers going up from 1 to n/2 and then back down to 1 again
spatialIndexes = [1:(n/2-1), (n/2):-1:1];

% create low-pass apodization window
switch lower(type) % to make it case-insensitive
    case 'cosine'
        apodize = cos((pi*spatialIndexes./cutoff)/(n)); % cosine window
        apodize = [1, apodize];
        ind_cut = find(apodize<0);
        if ~isempty(ind_cut)
            apodize(ind_cut(1):ind_cut(end)) = 0; %cutoff
        end
        clear ind_cut
    case 'hamming'
        apodize = 0.54 - 0.46 .* cos(2 .* pi .* linspace(n/2 - n/(2*cutoff), n/2 + n/(2*cutoff), n+1) ./ n);
        if cutoff < 1
            setToZero = floor(n/2 - cutoff * n/2);
            apodize(1:setToZero) = 0;
        end
        apodize(end-n/2+1:end) = apodize(n/2:-1:1); % make sure the window is symmetric (no rounding error)
        apodize(end) = []; % deleting last value, so that the window is periodic
        apodize = ifftshift(apodize); % shift the window so that the DC term is the first term
    case 'hann'
        apodize = 0.5 .* (1 - cos(2 .* pi .* linspace(n/2 - n/(2*cutoff), n/2 + n/(2*cutoff), n+1) ./ n));
        if cutoff < 1
            setToZero = floor(n/2 - cutoff * n/2);
            apodize(1:setToZero) = 0;
        end
        apodize(end-n/2+1:end) = apodize(n/2:-1:1); % make sure the window is symmetric (no rounding error)
        apodize(end) = []; % deleting last value, so that the window is periodic
        apodize = ifftshift(apodize); % shift the window so that the DC term is the first term
    case 'shepp-logan'
        apodize = sinc(linspace(-1/pi,1/pi,n+1));
        if cutoff < 1
            setToZero = floor(n/2 - cutoff * n/2);
            apodize(1:setToZero) = 0;
        end
        apodize(end-n/2+1:end) = apodize(n/2:-1:1); % make sure the window is symmetric (no rounding error)
        apodize(end) = []; % deleting last value, so that the window is periodic
        apodize = ifftshift(apodize); % shift the window so that the DC term is the first term
    case 'ram-lak'
        apodize = ones(1,n-1);
        apodize(spatialIndexes >= ((n/2) * cutoff)) = 0;
        apodize = [1, apodize];
    case 'none'
        apodize = ones(1,n);
end

% TODO - make apoOnly false by default
switch apoOnly
    case true
        kernel = apodize;
    case false
        if ~strcmpi(type, 'none') % if 'none' wasn't chosen
            % Create a ram-lak spatial kernel, transform into frequency domain
            imp_char = zeros(1, n-1);
            imp_char(1:2:end) = -1 ./ ( (spatialIndexes(1:2:end) .* pi).^2);
            imp_char = [0.25 imp_char];
            ramlak = fft(imp_char)./2;
            % multiply low-pass window with ram-lak kernel
            kernel = ramlak .* apodize;
        else % if 'none' was chosen
            % currently, there are only 1s in apodize, so output just this
            kernel = apodize;
        end
end

% % Visualize
% plot(abs(fftshift(kernel)))

end

