function [reconLog, saveMem] = reconstruct_ASTRA_FDK(data)
%UNTITLED11 Summary of this function goes here
%   handles reconstruction 
%   Detailed explanation goes here

try % wrap in try/catch, just to make sure I delete any temp file I create before exiting the function
    
    % get astra version, write into log
    ASTRAversion = astra_mex('version');
    % create log text
    reconLog = sprintf('%s %i\n', 'Astra version:', ASTRAversion);


    % % first try running a basic astra test to see if everything works
    % astra_test

    % parse the reconstruction settings in recSet
    [r, inputLog] = parseReconInputs(data.settings);
    
    
    % write to log
    reconLog = sprintf('%s%s\n', reconLog, inputLog);
    % TODO - do this in an easier way? (also needs to be changed in reconstructionReport then)

    % Fix in case ASTRA crashes: Limit GPU memory ASTRA uses, or calculations
    % take too long and hit a watchdog timer in Windows, which causes an error
    % See: https://github.com/astra-toolbox/astra-toolbox/issues/79
    if r.limitGPU
        mem_limit = 1500e6; % TODO - adjust this value
        astra_mex('set_gpu_index', 0, 'memory', mem_limit);
        % write into log
        reconLog = sprintf('%s%s %e %s\n', reconLog, 'GPU memory used by ASTRA limited to', mem_limit, 'bytes to prevent crashes');
    else
        astra_mex('set_gpu_index', 0, 'memory', inf);
        % write into log
        reconLog = sprintf('%s%s %e %s\n', reconLog, 'GPU memory used by ASTRA not limited');
    end

    % put geom in temporary variable - to make code more readable
    geom = copy(data.geom);

    % if using astra 1.9 or newer, create custom filter to pass to FDK
    if ASTRAversion >= 109
        % Get filter kernel
        kernel = filterKernel(geom.width, r.filterType, r.freqCutoff, false);
        % ASTRA expects only half the kernel
        halfFilterSize = floor(length(kernel) / 2) + 1;
        kernel = kernel(1:halfFilterSize);
        % repmat to number of projections
        kernel = repmat(kernel, geom.num_proj, 1);
        % create filter geometry
        filter_geom = astra_create_proj_geom('fanflat', geom.cb_mag, ...
            halfFilterSize, geom.angles, ...
            geom.SOD/geom.vox_size, ...
            geom.ODD/geom.vox_size);
        % create filter object
        filter_id = astra_mex_data2d('create', '-sino', filter_geom, kernel);
        % TODO - make a subfunction for this?
    else % else perform LP filtering outside ASTRA
        olderAstraApodize(data, r.filterType, r.freqCutoff);
        % write into log
        reconLog = sprintf('%s%s\n', reconLog, 'ASTRA version is older than 1.9 - low pass filtering done outside FDK');
    end
    
    
    % calculate memory demands for reconstruction - but not for the preview
    if isfield(data.misc, 'preview') && data.misc.preview == false % TODO - make sure I change this if I change it in the projectionData object
        [batch_size, saveMem] = memDemands(data);
    else
        batch_size = data.geom.height;
        saveMem = 0;
    end
    
    
    % write into log, create temporary file in temporary directory
    if saveMem
        % create temp file, set permission for appending to end of file
        tempFileName = tempname; % TODO - add some readable part of the name? for easier identification just in case
        tempFileID = fopen(tempFileName, 'a'); % TODO - use A istead of a? (flushing output buffer)

        reconLog = sprintf('%s%s\n%s\n', reconLog, 'Insufficient memory to reconstruct entire volume', 'Trying reconstruction in smaller batches');
    else
        reconLog = sprintf('%s%s\n%s\n', reconLog, 'Memory sufficient', 'Reconstructing entire volume');
    end


    % determine the amount of slices reconstructed at once
    slice_ranges = 1 : batch_size : geom.reconHeight;
    if saveMem
        reconLog = sprintf('%s%s %i %s\n', reconLog, 'Reconstructing tomograms in batches of', batch_size, 'slices');
    else
        reconLog = sprintf('%s%s\n', reconLog, 'Reconstructing all tomogram slices at once');
    end


    % set up projection geometry
    proj_geom = astra_create_proj_geom('cone', geom.cb_mag, geom.cb_mag, ...
        geom.height, geom.width, geom.angles, ...
        geom.SOD/geom.vox_size, geom.ODD/geom.vox_size);
    % shift projections if CoR is not centered
    if abs(data.geom.CoR_shift) %> 0.5 % TODO - should I keep > 0.5?
        if ASTRAversion >= 109
             proj_geom = astra_geom_postalignment(proj_geom, data.geom.CoR_shift);
        else
            shiftProjs(data);
        end
    end
    % create projection data object
    proj = data.proj;
    data.proj = [];
    proj = permute(proj, [2, 3, 1]); % so that projs are the way ASTRA likes it
    proj_id = astra_mex_data3d('create','-proj3d', proj_geom, proj); % link removed from 1.9 :(, have to use create
    clear proj % delete projections from Matlab to save RAM
    
    % set up basic algorithm settings
    cfg = astra_struct('FDK_CUDA');
    % link projection data object with algorithm
    cfg.ProjectionDataId = proj_id;
    if ASTRAversion >= 109 % if using newer ASTRA, link custom filter w/ algo
        % link custom filter object with algorithm
        % cfg.option.FilterType = 'projection'; % apparently this is not used
        cfg.option.FilterSinogramId = filter_id;
    end
    if data.geom.shortscanFlag
        cfg.option.ShortScan = 1;
    end
    % cfg.option.VoxelSuperSampling = 0; % TODO - study up on voxel supersampling

    for k = slice_ranges % taking a subset/batch of slices to reconstruct

        % get Z indexes of range to be reconstructed
        if k ~= slice_ranges(end) % if it's not the last batch
            sliceStart = k; % index of 1st slice of current batch
            sliceEnd = k + batch_size - 1; % index of last slice of current batch
        else
            sliceStart = k; % 1st slice of batch
            sliceEnd = geom.reconHeight; % if it's the last batch, just take whatever is left
        end
        % number of slices in batch (equal to batch_size, except last batch)
        sliceNum = sliceEnd + 1 - sliceStart;

        % set up volume geometry
        min_z = -geom.reconHeight/2 + sliceStart - 1; % lower z limit
        max_z = -geom.reconHeight/2 + sliceEnd; % upper z limit
        lim_xy = geom.reconWidth/2; % limits in x and y directions
        vol_geom = astra_create_vol_geom(...
            geom.reconWidth, geom.reconWidth, sliceNum, ... % size
            -lim_xy, lim_xy, -lim_xy, lim_xy, min_z, max_z); % limits

        % create reconstruction data object with current geometry
        rec_id = astra_mex_data3d('create', '-vol', vol_geom);
        
        % link reconstruction data object with algorithm
        cfg.ReconstructionDataId = rec_id;

        % setup and run algorithm
        alg_id = astra_mex_algorithm('create', cfg);
        astra_mex_algorithm('run', alg_id);
        % clean up algorithm
        astra_mex_algorithm('delete', alg_id);

        % Get the result
        tomograms = astra_mex_data3d('get_single', rec_id);
        % clean up reconstruction data
        astra_mex_data3d('delete', rec_id);

        % get minimum and maximum value in reconstruction
        if k == slice_ranges(1) % if it is the first batch to be reconstructed
            % just get the min and max
            rec_minimum = min(tomograms(:));
            rec_maximum = max(tomograms(:));
        else
            % get min and max of current batch, compare with previous values
            current_min = min(tomograms(:));
            current_max = max(tomograms(:));
            if current_min < rec_minimum; rec_minimum = current_min; end
            if current_max < rec_maximum; rec_maximum = current_max; end
        end

        if saveMem % if reconstruction is bigger than what the RAM can hold
            % write it to the created temp file
            fwrite(tempFileID, tomograms, 'single');
            clear tomogams % clear tomograms from matlab to save memory
            if k == slice_ranges(end) % close file after last batch is done
                fclose(tempFileID);
            end
        end
        % write info about reconstructing current batch to log
        reconLog = sprintf('%s%s %i %s %i\n', reconLog, 'Successfully reconstructed slices', sliceStart, 'to', sliceEnd);
    end
    
    % delete projections altogether if saveMem is true - to clear space
    if saveMem
        % clean up of projection data object
        astra_mex_data3d('delete', proj_id);
    else % else return them to the projectionData object
        % get projection data back from ASTRA
        proj = astra_mex_data3d('get_single', proj_id);
        % clean up of projection data object
        astra_mex_data3d('delete', proj_id);
        % permute data back to projection view
        proj = permute(proj, [3, 1, 2]);
        % put back into projectionData object
        data.proj = proj;
        % clear proj variable
        clear proj
    end
    
    % if saveMem is true, load reconstruction from temporary file
    if saveMem
        % create a memory map to the file
        tempMap = memmapfile(tempFileName, 'Format', {'single', ...
            [geom.reconWidth, geom.reconWidth, geom.reconHeight], 'x'}, ...
            'Writable', true);
        
        % map tomograms onto a 16-bit range
        tempMap.Data.x = (2^16 - 1) .* ...
                (tempMap.Data.x - rec_minimum) ./ (rec_maximum - rec_minimum);
        
        data.tomo = tempMap.Data.x; % copy to data object
        clear tempMap % clear memory map object from memory
        % NB: data will be copied to tomograms var when tempMap is cleared
        
        % finally delete the temporary file 
        delete(tempFileName);
        
    else
        % map tomograms onto a 16-bit range, do not convert yet
        tomograms = (2^16 - 1) .* ...
                (tomograms - rec_minimum) ./ (rec_maximum - rec_minimum);
            
        % write tomograms into object
        data.tomo = tomograms;
        clear tomograms % clear from memory
        
    end
    reconLog = sprintf('%s%s\n', reconLog, 'Tomogram values mapped onto range 0 to (2^16 - 1)');
    
    % cyllinder trim the tomograms
    if ~saveMem
        cyllinderTrim(data, mean(data.tomo(:)));
        % write to log
        reconLog = sprintf('%s%s\n', reconLog, 'Values outside FoV cyllinder trimmed');
    else
        reconLog = sprintf('%s%s\n', reconLog, 'Values outside FoV cyllinder not trimmed due to memory constraints');
    end
    
    % attempt to wipe all astra data to prevent memory leaks
    astra_clear
    
catch err
    
    % delete temp file if it exists
    if exist('tempFileID', 'var')
        fclose('all'); % just close all open files so no errors are thrown
        if ~isempty( which(tempFileName) ) % if temp file still exists
            delete(tempFileName); % then delete it
        end
    end
    % attempt to wipe all astra data to prevent memory leaks
    astra_clear;
    % finally throw original error (rethrow traces back to original source)
    rethrow(err);
    
end

end


function [batch_size, saveMem] = memDemands(data)
% automatically determine batch_size (how many slices are reconstructed
% in one go) precaution to not run out of RAM.

geom = data.geom;

% Calculate amount of bytes required for reconstruction volume
memneeded = geom.reconWidth .^2 * geom.reconHeight * 4; % bytes
% Get available system memory
[~,sys] = memory;
% TODO - call to memory takes quite long (~2s) - is there a better way?
memavailable = sys.PhysicalMemory.Available; % currently available memory
% check if there's enough memory for projections AND tomograms
if memavailable < memneeded % if there isn't enough memory
    % if there is not enough memory for both projections and tomograms
    max_batch = floor(geom.reconHeight * (memavailable/memneeded) * 0.5);
    % times 0.5 just to be sure there's enough room for
    % moving projections between ASTRA and Matlab
    saveMem = true;
else % if there is enough memory
    max_batch = geom.reconHeight;
    saveMem = false;
end
batch_size = min(geom.reconHeight, max_batch);
% TODO - check this method of determining batch_size, check the 0.5 in max_batch
end


function olderAstraApodize(data, filterType, freqCutoff)
%if the ASTRA version used is older than 1.9, a LP filter is applied to
%projection data before it is passed to the reconstruction algorithm
%   Detailed explanation goes here

% get the filter kernel
kernel = filterKernel(data.geom.width, filterType, freqCutoff, true);
% last argument specifies whether to output only apodization window

% replicate kernel to projection size
kernel = repmat(kernel,data.geom.height,1);
% putting width in separate variable to reduce overhead in parfor
width = data.geom.width; 
% temporarily put proj outside the object, so it can go to parfor
proj = data.proj; 
data.proj = [];
parfor p = 1:data.geom.num_proj % filtering individual projections
    % fft projection, multiply w/ kernel, ifft
    p_filt = fft(proj(:,:,p), size(kernel, 2), 2) .* kernel;
    p_filt = real( ifft(p_filt, [], 2));
    % trim back to original width, put back into array
    p_filt = p_filt(:, 1:width);
    proj(:,:,p) = p_filt;
end
% put projections back into object
data.proj = proj;
end


function [r, inputLog] = parseReconInputs(recSet)
% use inputParser for algorithm-specific settings

p = inputParser;
validfilters = {'cosine', 'hamming', 'hann', 'shepp-logan', 'ram-lak', 'none'}; % TODO - expand if I add more filters
checkfilter = @(x) any(validatestring(x, validfilters));
addParameter(p, 'filterType', 'cosine', checkfilter)
addParameter(p, 'freqCutoff', 0.85, @(x) isnumeric(x) && isscalar(x) && x>0)
addParameter(p, 'limitGPU', false, @(x) isscalar(x) && (islogical(x) || isnumeric(x)) )
p.KeepUnmatched = true; % no error when extra settings are supplied
% NB: but will throw error when e.g. filterType doesn't match a valid option
parse(p,recSet.rec.algoSettings)
% parsed inputs are in p.Results, get them out - into 'r' struct
r = p.Results;

% make a report about the inputs
inputLog = sprintf('%s\n', 'Input parameters:');
% write used inputs into log
inputCell = cellfun( @num2str, [fieldnames(p.Results), struct2cell(p.Results)], 'UniformOutput', 0);
for k = 1:size(inputCell, 1)
    inputLog = sprintf('%s\t%s%s %s\n', inputLog, inputCell{k, 1}, ':', inputCell{k, 2});
end
inputLog = sprintf('%s%s\n', inputLog, 'Unused inputs:');
% write unused inputs into log
unusedCell = cellfun( @num2str, [fieldnames(p.Unmatched), struct2cell(p.Unmatched)], 'UniformOutput', 0);
if size(unusedCell, 1) > 0
    for k = 1:size(unusedCell, 1)
        inputLog = sprintf('%s\t%s%s %s\n', inputLog, unusedCell{k, 1}, ':', unusedCell{k, 2});
    end
else
    inputLog = sprintf('%s\t%s\n', inputLog, 'none');
end
% TODO - wrap in try/catch?
end


