function [data, phantomStruct] = makeSyntheticSino(ph, varargin)
%takes synthetic 2D data as input, outputs sinogram in PROJECTIONDATA obj.
%   MAKESYNTHETICOFFSETSINO takes as input a 2D matrix representing a
%   synthetic tomographic phantom, along with a structure specifying how
%   this phantom should be transformed into projection data. it outputs a
%   PROJECTIONDATA object containing the projection data of the input
%   phantom. It uses the ASTRA toolbox to create transform the phantom.
%   NB: this function is meant to simulate acquisition by the Nano3DX
%   Inputs:
%       ph   - 2D matrix (single, double) - the size should correspond
%                   with the size the data should be at binning 1
%            OR
%              a character vector specifying the phantom to be used (see
%              SPITOUTPHANTOM for options)
%       set - structure of various settings specifying how the
%           phantom should be transformed into projectiondata. A list of
%           these settings is as follows:
%               set.det_count - number of detector bins AFTER BINNING
%               set.binning - binning of detector pixels
%               set.extension - extension of the field of view by offsetting the sample
%               set.pix_size_b1 - pixel pitch at binning 1 (micrometers)
%               set.SOD - source-origin distance (micrometers)
%               set.ODD - origin-detector distance (micrometers)
%               set.num_proj - number of projections (positive scalar integer)
%               set.incident_flux - incident x-ray flux (positive scalar integer)
%               set.poissonNoise - logical flag whether to add poisson noise
%               set.gaussianNoise - amount of gaussian noise to be added (standard deviation)
%               set.impulseNoise - ratio of pixels impacted by impulse noise
%               set.rings - standard deviation of ring artifacts to be added
%               set.exp_decrease - an exponential decrease in the gray levels in projections as a function of the projection index (<=1, amount by which the gray values decrease in the last projection when compared to the first)
%               set.transmissionOutput - logical flag whether to output
%                   transmission data (as opposed to absorption data)
%               set.fullSino - logical flag whether output with extension
%                   should be full sinogram or truncated
%               set.angleRange - angular range of projections (2*pi default)
%               set.phantomOption - used only if ph is a string
%               set.outputBinnedPhantom - output a version of the phantom
%               with pixel binning applied in the output structure
% 
%   Outputs
%       data - PROJECTIONDATA object (see help projectionData)
%       phantomStruct - a structure containing information about the
%                       phantom, including location of the CoR, size (N), a
%                       matrix containing the original phantom, and, if 
%                       set.outputBinnedPhantom is set to true, a binned
%                       version of the phantom
%   Example:
%       set = struct; % set up settings structure
%       set.det_count = 1648; 
%       set.binning = 2; 
%       set.extension = 1.8;
%       set.pix_size_b1 = 0.54; 
%       set.SOD = 275000;
%       set.ODD = 5000;
%       set.num_proj = 1600;
%       set.incident_flux = 5e5;
%       set.poissonNoise = false;
%       set.gaussianNoise = 0;
%       set.impulseNoise = 0;
%       set.rings = 0;
%       set.exp_decrease = 1;
%       set.transmissionOutput = true;
%       set.fullSino = false;
%       set.angleRange = 2*pi;
%       set.phantomOption = 1;
%       set.outputBinnedPhantom = true;
%       
%       [data, phantomStruct] = makeSyntheticOffsetSino(P, set); % run the function

%% check inputs, create phantom if needed

s = parseInputs(varargin{:}); % parse inputs
% NB: in s, pix_size is already after binning

N = round(s.det_count * s.extension); % true size

if ischar(ph)
    ph = spitOutPhantom(N*s.binning, ph, s.phantomOption); % NB: times binning
else
    validateattributes(ph, {'single', 'double'}, {'2d', 'square'}, ...
                            'makeSyntheticOffsetSino', 'ph', 1)
end


%% create sinogram
volume_width = size(ph, 1) / s.binning; % should be a square

% Set scan parameters
cb_mag = (s.SOD+s.ODD) / s.SOD; % calculate cone beam magnification
vox_size = s.pix_size / cb_mag; % calculate voxel size (already after binning)

angles = linspace(0, s.angleRange, s.num_proj + 1); % calculate projection angles (from 0, in radians)
% TODO - make the range flexible

angles(end) = []; % get rid of last angle (projection at 0 is the same as at 2*pi)

% Create sinogram
height = 1;
if s.fullSino
%     proj_geom = astra_create_proj_geom('fanflat', cb_mag, volume_width, angles, s.SOD/vox_size, s.ODD/vox_size);
    proj_geom = astra_create_proj_geom('cone', ...
        cb_mag * s.binning, cb_mag * s.binning, ...
        height, volume_width, angles, ...
        (s.SOD/vox_size) * s.binning, (s.ODD/vox_size) * s.binning );
else % else create offset-sample sinogram
%     proj_geom = astra_create_proj_geom('fanflat', cb_mag, s.det_count, angles, s.SOD/vox_size, s.ODD/vox_size);
    proj_geom = astra_create_proj_geom('cone', ...
        cb_mag * s.binning, cb_mag * s.binning, ...
        height, s.det_count, angles, ...
        (s.SOD/vox_size) * s.binning, (s.ODD/vox_size) * s.binning );
    
    % shift in CoR - by the desired amount 'extension'
    shift =  s.binning * (volume_width - s.det_count) / 2; % times binning cos the volume can have a finer resolution
    % old
%     shift = s.binning * s.det_count * (s.extension/2 - 0.5); % will shift the detector by an offset that will extend the FoV by the desired amount 'extension'
    
    % convert projection geometry to a vector geometry
    proj_geom = astra_geom_2vec(proj_geom);
    %   ( srcX, srcY, srcZ, dX, dY, dZ, uX, uY, uZ, vX, vY, vZ )
    %   I have to shift the source and detector along the X-axis - srcX, srcY, dX, dY
    proj_geom.Vectors(:, 1) = proj_geom.Vectors(:, 1) + shift .* cos(angles)'; % shift source in X axis
    proj_geom.Vectors(:, 2) = proj_geom.Vectors(:, 2) + shift .* sin(angles)'; % shift source in Y axis
    proj_geom.Vectors(:, 4) = proj_geom.Vectors(:, 4) + shift .* cos(angles)'; % shift detector in X axis
    proj_geom.Vectors(:, 5) = proj_geom.Vectors(:, 5) + shift .* sin(angles)'; % shift detector in Y axis
end

% vol_geom = astra_create_vol_geom( volume_width, volume_width);
vol_geom = astra_create_vol_geom( volume_width * s.binning, volume_width * s.binning, height);

% [~, sino] = astra_create_sino_cuda(ph, proj_geom, vol_geom);
[~, sino] = astra_create_sino3d_cuda( double(ph), proj_geom, vol_geom ); % converting ph to double - doesn't work w/ single for some reason

% normalize sino
sino = ( sino - min( sino(:) ) ) / ( max( sino(:) ) - min( sino(:) ) );

%% process sinogram - add noise etc.

% create transmission sinogram
I = s.incident_flux .* exp( -sino );

if s.exp_decrease ~= 1 % add decrease in grayscale values
    exp_end = 2;
    decrease = exp( linspace( 0, -exp_end, numel(angles) ) ); % get an exponential between 0 and 2 % TODO - change this?
    ratio = (s.exp_decrease - 1) / ( exp(-exp_end) - 1 ); % ratio of change in the case the end is s.exp_decrease* the start, and the case where it is exp(-exp_end)* the start

    decrease = decrease .* ratio; % making it so that decrease(end) is decrease(1)*s.exp_decrease
    decrease = decrease + (1 - decrease(1)); % raise the baseline so that decrease(1) == 1
    
    I = I .* repmat(decrease(:)', size(I, 1), 1); % multiply the output by the given exponential function
end

if s.rings
    I = I + s.incident_flux .* s.rings .* randn(size(I, 1), 1);
end

if s.poissonNoise % add poisson noise
    I = poissrnd(I);
end

if s.gaussianNoise
    I = I + s.gaussianNoise.*randn(size(I));
end

if s.impulseNoise
    spots = rand(size(I)) < s.impulseNoise; % get locations of noise
    indices = find(spots); % get indices of these locations
    indices = indices( randperm( length(indices) ) ); % shuffle
    
    I(indices(1:round(length(indices)/2) )) = min(I(:)); % half are low
    I(indices(round( length(indices)/2 + 1):end )) = max(I(:)); % half are high
    
end

if s.transmissionOutput
    sino = I;
else
    % get back noisy absorption sinogram
    sino = -log(I ./ s.incident_flux);
end

% calculate shift in mm
delta_T = (s.pix_size * (s.extension-1) * s.det_count/2) / 1000;

%% create outputs

% Create projectionData object
data = projectionData;
% permute sinogram and write into proj property of data
data.proj = permute(sino, [3, 1, 2]);
% input some metadata
data.editParams('FlatFieldCoefficient', s.incident_flux, ...
                'binning', s.binning, ...
                'delta_T', delta_T)
% edit scan geometry
data.geom.edit('width', s.det_count, ...
              'height', 1, ...
              'pix_size', s.pix_size, ...
              'SOD', s.SOD, ...
              'ODD', s.ODD, ...
              'angles', angles, ...
              'CoR', (s.det_count+1)/2, ...
              'reconWidth', s.det_count, ...
              'reconHeight', 1, ...
              'offsetFlag', true)
if s.angleRange > pi && s.angleRange < 2*pi
    data.geom.edit('shortscanFlag', true)
end

if s.fullSino
    data.geom.edit('width', volume_width, 'reconWidth', volume_width, ...
                   'CoR', (volume_width+1)/2, 'offsetFlag', false)
end

phantomStruct.N = N;
phantomStruct.CoR = (2*s.det_count - N + 1) / 2;
phantomStruct.gTruth = ph;
if s.outputBinnedPhantom
    phantomStruct.binnedPhantom = imresize(ph, 1/s.binning, 'nearest');
end

end

function [ s ] = parseInputs(varargin)

% parse inputs
p = inputParser;
addParameter(p, 'det_count', 1024, @(x) isnumeric(x) && isscalar(x) && x>=1);
addParameter(p, 'extension', 1, @(x) isnumeric(x) && isscalar(x) && x>=1);
addParameter(p, 'pix_size_b1', 1, @(x) isnumeric(x) && isscalar(x) && x>0);
addParameter(p, 'binning', 1, @(x) isnumeric(x) && isscalar(x) && x>0 && ~mod(x,1));
addParameter(p, 'SOD', 5000, @(x) isnumeric(x) && isscalar(x) && x>=0);
addParameter(p, 'ODD', 1000, @(x) isnumeric(x) && isscalar(x) && x>=0);
addParameter(p, 'num_proj', 360, @(x) isnumeric(x) && isscalar(x) && x>=0 && ~mod(x,1));
addParameter(p, 'incident_flux', 1e5, @(x) isnumeric(x) && isscalar(x) && x>=0 && ~mod(x,1));
addParameter(p, 'poissonNoise', false, @(x) isnumeric(x) || islogical(x) && isscalar(x));
addParameter(p, 'gaussianNoise', 0, @(x) isnumeric(x) && isscalar(x) && x>=0);
addParameter(p, 'impulseNoise', 0, @(x) isnumeric(x) && isscalar(x) && x>=0 && x<=1);
addParameter(p, 'rings', 0, @(x) isnumeric(x) && isscalar(x) && x>=0);
addParameter(p, 'exp_decrease', 1, @(x) isnumeric(x) && isscalar(x) && x>=0 && x<=1);
addParameter(p, 'transmissionOutput', true, @(x) isnumeric(x) || islogical(x) && isscalar(x));
addParameter(p, 'fullSino', false, @(x) isnumeric(x) || islogical(x) && isscalar(x));
addParameter(p, 'angleRange', 2*pi, @(x) isnumeric(x) && isscalar(x) && x>=0 && x<=2*pi);

addParameter(p, 'phantomOption', 1, @(x) isnumeric(x) && isscalar(x) && x>=0 && mod(x,1)==0);
addParameter(p, 'outputBinnedPhantom', false, @(x) isnumeric(x) || islogical(x));

p.KeepUnmatched = true;

parse(p, varargin{:});

s.det_count = p.Results.det_count;
s.extension =  p.Results.extension;
s.pix_size =  p.Results.pix_size_b1 .* p.Results.binning; % after binning!
s.binning =  p.Results.binning;
s.SOD =  p.Results.SOD;
s.ODD =  p.Results.ODD;
s.num_proj =  p.Results.num_proj;
s.incident_flux =  p.Results.incident_flux;
s.poissonNoise =  p.Results.poissonNoise;
s.gaussianNoise =  p.Results.gaussianNoise;
s.impulseNoise =  p.Results.impulseNoise;
s.rings = p.Results.rings;
s.exp_decrease = p.Results.exp_decrease;
s.transmissionOutput = p.Results.transmissionOutput;
s.fullSino = p.Results.fullSino;
s.angleRange = p.Results.angleRange;
s.phantomOption = p.Results.phantomOption;
s.outputBinnedPhantom = p.Results.outputBinnedPhantom;
end

