function P = spitOutPhantom(N, name, thirdArg)
% function that outputs predefined phantoms
% depends on the Matlab function PHANTOM

% 'shepp-logan', 'original shepp', 'siemens star', 'circle', 'concentric parabolic'
% 'concentric linear', 'circles', 'point', 'grid', 'squares'

% thirdArg can be supplied for some functions

% TODO - add better description

validateattributes(N, {'numeric'}, {'scalar', 'integer', 'positive'}, 'spitOutPhantom', 'N', 1);
validateattributes(name, {'char'}, {'scalartext'}, 'spitOutPhantom', 'name', 2);

% TODO - throw phantom-specific errors

name = lower(name);

switch name
    case 'shepp-logan'
        P = phantom(N);
    case 'original shepp'
        P = phantom('Shepp-Logan', N);
    case 'siemens star'
        validateattributes(thirdArg, {'numeric'}, {'scalar', 'integer', 'positive'}, 'spitOutPhantom', 'thirdArg', 3);
        P = getSiemensStar(N, thirdArg); % number of rays
    case 'circle'
        % TODO - add radius as thirdArg?
        P = phantom([1, 0.9, 0.9, 0, 0, 0], N);
    case 'concentric parabolic'
        validateattributes(thirdArg, {'numeric'}, {'scalar', 'integer', 'positive'}, 'spitOutPhantom', 'thirdArg', 3);
        P = getConcentricRingsParabolic(N, thirdArg); % number of circles
    case 'concentric linear'
        validateattributes(thirdArg, {'numeric'}, {'scalar', 'integer', 'positive'}, 'spitOutPhantom', 'thirdArg', 3);
        P = getConcentricRingsLinear(N, thirdArg); % number of circles
    case 'circles'
        validateattributes(thirdArg, {'numeric'}, {'scalar', 'integer', 'positive'}, 'spitOutPhantom', 'thirdArg', 3);
        P = getCircles(N, thirdArg); % number of circles
    case 'point'
        P = getPoint(N);
    case 'grid'
        validateattributes(thirdArg, {'numeric'}, {'scalar', 'integer', 'positive'}, 'spitOutPhantom', 'thirdArg', 3);
        P = getGrid(N, thirdArg); % size of holes
    case 'squares'
        validateattributes(thirdArg, {'numeric'}, {'scalar', 'integer', 'positive'}, 'spitOutPhantom', 'thirdArg', 3);
        P = getSquares(N, thirdArg); % size of squares
    case 'resolution'
        validateattributes(thirdArg, {'numeric'}, {'scalar', 'integer', 'positive'}, 'spitOutPhantom', 'thirdArg', 3);
        P = getResolution(N, thirdArg);
    otherwise
        P = phantom(N);
end

P = single(P); % output single

end

function P = getSiemensStar(N, w)
% Olivier Lvque
% https://www.mathworks.com/matlabcentral/fileexchange/68982-siemens-star

li = linspace(-1, 1, N);
[X, Y] = meshgrid(li);
[th, rad] = cart2pol(X, Y);
P = 1 + sin(w*th);
P = double((P>1)&(rad<0.9));
end

function P = getConcentricRingsParabolic(N, num_rings)

li = linspace(-sqrt(num_rings*2), sqrt(num_rings*2), N);
[X, Y] = meshgrid(li);
parabola = X.^2 + Y.^2; % create paraboloid
parabola = floor(parabola); % round all values down to whole numbers

P = zeros(N);
for k = 0:2:(num_rings*2 - 1)
    P(parabola==k) = 1;
end

end

function P = getConcentricRingsLinear(N, num_rings)

li = linspace(-num_rings*2, num_rings*2, N);
[X, Y] = meshgrid(li);
[~,rho] = cart2pol(X, Y);
rho = floor(rho);

P = zeros(N);
for k = 0:2:(num_rings*2-1)
    P(rho==k) = 1;
end
end

function P = getCircles(N, num_circles)

radius = 0.25;
radii = repmat(radius, num_circles, 1);

angle_to_middle = linspace(0, 2*pi, num_circles + 1)';
angle_to_middle(end) = [];
x_coord = 0.6 .* cos(angle_to_middle);
y_coord = 0.6 .* sin(angle_to_middle);

P = phantom([ones(num_circles, 1), radii, radii, ...
             x_coord, y_coord, zeros(num_circles, 1)], N);
end

function P = getPoint(N)

    P = zeros(N);
    
    half = (N+1)/2;
    if mod(N, 2) % if odd, set middle pixel to 1
        P(half, half) = 1;
    else % if even - set middle 4 pixels to 1
        P(floor(half):ceil(half), floor(half):ceil(half)) = 1;
    end
end

function P = getGrid(N, thickness)
    
    P = zeros(N);
    
    for k = 1:(2*thickness):N
        P(k:(k+thickness-1), :) = 1;
        P(:, k:(k+thickness-1)) = 1;
    end
      
    [X, Y] = meshgrid( linspace(-1, 1, N) );
    mask = (X.^2 + Y.^2) > 0.95.^2;
    
    P = P(1:N, 1:N); % make sure P is the right size
    
    P(mask) = 0;
    
    
    
end

function P = getSquares(N, sizeGrid)
    % sizeGrid is the width of individual squares
    P = zeros(N);
    
    for k = 1:(2*sizeGrid):N
        P(k:(k+sizeGrid-1), :) = 1;
    end
    for k = 1:(2*sizeGrid):N
        P(:, k:(k+sizeGrid-1)) = 0;
    end
    
    [X, Y] = meshgrid( linspace(-1, 1, N) );
    mask = (X.^2 + Y.^2) > 0.95.^2;
    
    P = P(1:N, 1:N); % make sure P is the right size
    
    P(mask) = 0;
    
end

function P = getResolution(N, scale)
    
    P = zeros(N);
    
    middle = N/2;

    R = (N*0.8)/2;
    
    basePattern = repmat([1, 0, 1, 0, 1, 0, 1, 0, 1, 0], round(N/20), 1);
    
%     patternWidths = [1, 2, 3, 4, 5, 6, 7];
%     patternWidths = [4, 3, 2, 1];
%     patternWidths = [1, 2, 3, 4];
    patternWidths = [0, 1, 2, 3, 4, 3, 2, 1];
%     patternWidths = [0, 4, 3, 2, 1, 2, 3, 4];
    numPatterns = numel(patternWidths);
    
    for k = 0:(numPatterns-1)
        
        pattern = repelem(basePattern, 1, scale * patternWidths(k+1));
        
%         currAngle = 2*pi*(k/numPatterns);
        currAngle = pi*(k/numPatterns);
        
%         pattern = imrotate(pattern, 180*currAngle/pi);
        
        patternHeight = size(pattern, 1);
        patternWidth = size(pattern, 2);
        
        coords = [ floor( middle + R*sin( currAngle ) - patternHeight/2), ...
                   floor( middle + R*cos( currAngle ) - patternWidth/2) ];
        
        P( coords(1):(coords(1)+patternHeight-1), ...
           coords(2):(coords(2)+patternWidth-1) ) = pattern;
        
    end
    
end

